/* draw.c - (C) Richard K. Lloyd 2010 <webmaster@lottery.merseyworld.com>

   Simulate a realistic spread of ticket purchases given a total
   ticket sales figure. Then calculate the number of winners/prize
   amounts in each category as fast as possible after each ball has been drawn.
   This basically proves that the live UK lottery TV show *can*
   display the prizes/winners in real time, which they've never done.
*/

/* Global definitions */
#include "draw.h"

/* Yes, I actually include the other *.c source files in this one - it shaves
   10% off the execution time by doing so, because gcc can better optimise
   a single source file than multiple ones. You can compile and link them
   separately if you insist, but why lose performance doing so? */
#ifdef WHOLE_SOURCE
#include "generate.c"
#include "gui.c"
#include "live.c"
#endif

#ifdef MULTITHREAD
#ifdef WHOLE_SOURCE
#include "pthread.c"
#endif

/* Number of combinations in each "slice" of drawn from upper */
int combo_slice[MAX_UPPER+1];

/* Combination speed arrays from live.c */
#ifdef VARIABLE_COMBO_CHUNK
extern int speed_combo[MAX_DRAWN-1];
#endif
extern int scaled_speed_combo[MAX_DRAWN-1];
#endif

/* Holds counter (2 bytes per comb) of all combs */
static unsigned char *tickets=(unsigned char *)NULL;

/* Factorial array used when calculating combinations. It reaches MAX_UPPER!,
   so it needs to be a floating-point array */
static double fact[MAX_UPPER+1];

/* The number of combinations of [0..drawn] from [0..upper] are stored
   in this array. */
long combs[MAX_UPPER+1][MAX_UPPER+1];

int combidx[MAX_DRAWN+2]; /* Used to store the current combination */

/* Initialise some variables, though several of them can't be overridden
   by user input or command-line options (well, not yet at least!) */
int min_match=MIN_MATCH,drawn=DEF_DRAWN,upper=DEF_UPPER,totdrawn=DEF_DRAWN+1;
long num_combs,ticket_sales=0,rollover=-1,topup=1;

/* All of these variables, though, are controlled through command-line flags */
int read_tickets=0,verify=0,timings=1,write_tickets=0,
    have_bonus_ball=1,numsofar=0,plusminus=0,show_losers=0,
    show_prize_early=0,show_all_prizes=0,csv_output=0,
    three_match_prize=THREE_MATCH_PRIZE_DEFAULT;
long randseed=-1;
int cores_used=1; /* Even with no multi-threading, still use 1 CPU core */

/* Functions found in other source files */
extern void feed_tickets_from_stdin(void),generate_tickets(void),
            perform_draw(void),show_verify_table(void),save_ball_set(void),
            restore_ball_set(void),request_gui(void);
extern char *mills(long num),*pounds(long num),*ask_for_string(char *prompt);
extern int insert_number(int done,int newnum);
extern int draw_from_file;
extern char drawn_file[BUFSIZ];
extern time_t draw_file_stamp;
extern long draw_file_size,ask_for_value(char *prompt);

/* Forward declarations to keep various compilers quiet */
int comb2idx();
int get_tickets_for_comb(int idx);
void put_tickets_for_comb(int idx,int numticks);

#ifdef MULTITHREAD
/* Multi-threading variables */

/* Get number of cores from /proc/cpuinfo on Linux or
   $NUMBER_OF_PROCESSORS on Windows */
int num_cores=0;

/* Found in pthread.c */
extern int wait_for_jobs;
extern int *thread_running;
extern void kill_all_children();
#endif /* MULTITHREAD */

void tidy_up()
{
   /* Last function called before exiting the program, so just free up
      the ticket memory array. Yes, I know then exiting will do this
      automatically, but this is just good housekeeping. */
   if (tickets!=(unsigned char *)NULL) (void)free((void *)tickets);

#ifdef MULTITHREAD
 /* Threads spawned, so tidy up nicely */
   if (cores_used>1) kill_all_children();
   if (thread_running!=(int *)NULL)
   {
      (void)free((void *)thread_running);
      thread_running=(int *)NULL;
   }
#endif
}

#ifdef MULTITHREAD
static void get_cpu_cores()
/* Work out how many CPU cores there are. Note that hyper-threaded CPU cores
   will be counted as whole CPU cores, despite only giving a small benefit,
   if any.

   Some of this code is from Blender (GPL 2 licensed, before you panic):
   https://svn.blender.org/svnroot/bf-blender/trunk/blender/source/blender/blenlib/intern/threads.c
*/
{
   int ztr;
#if defined(__linux)
   /* It's Linux, so there's actually a sysconf() call to
      get the number of CPU cores. Blender to the rescue, because
      I hated the naff /proc/cpuinfo parsing I was doing before. */
   num_cores=(int)sysconf(_SC_NPROCESSORS_ONLN);
#else
#if defined(__APPLE__) && defined(__MACH__)
   /* It's Mac OS X, so it's fiddly code lifted from Blender. */
   int mib[2];
   size_t len=sizeof(num_cores);

   mib[0]=CTL_HW; mib[1]=HW_NCPU;
   (void)sysctl(mib,2,&num_cores,&len,NULL,0);
#else
#ifdef __hpux
   /* It's HP-UX, but "man pstat" has a nice example that I'll
      use here. */
   struct pst_dynamic psd;
   if (pstat_getdynamic(&psd,sizeof(psd),(size_t)1,0)!=-1)
      num_cores=psd.psd_proc_cnt;
#else
   /* Surprisingly, Windows actually has the number of CPU cores
      handily stored in $NUMBER_OF_PROCESSORS (which the user
      can override and screw up royal style). Blender suggests
      this:

      SYSTEM_INFO info;
      GetSystemInfo(&info);
      t = (int) info.dwNumberOfProcessors;

      but GetSystemInfo() isn't available in Cygwin.
   */
   char *coreptr=getenv("NUMBER_OF_PROCESSORS");
   if (coreptr!=(char *)NULL) num_cores=atoi(coreptr);
#endif /* __hpux */
#endif /* __APPLE__ && __MACH__ */
#endif /* __linux */

   /* Blender limits the value to between 1 and the number of
      threads it can handle, so we do something similar... */

   /* If we couldn't work out the number of cores, just set to 1 */
   if (num_cores<1) num_cores=1;

   /* Also limit the number of cores to our maximum */
   else if (num_cores>MAX_CPU_CORES) num_cores=MAX_CPU_CORES;
#ifdef DEBUG
   (void)fprintf(stderr,"Number of CPU cores detected: %d\n",num_cores);
#endif

   /* Allocate memory for thread_running[] array */
   thread_running=(int *)malloc(num_cores*sizeof(int));
   if (thread_running==(int *)NULL)
   {
      (void)fprintf(stderr,"Unable to allocate memory for thread monitor - aborted\n");
      tidy_up();
      exit(1);
   }

   /* Indicate that there's no child threads running on any cores */
   numjobs=0;
   for (ztr=0;ztr<num_cores;ztr++) thread_running[ztr]=0;
}
#endif /* MULTITHREAD */

static long calc_combs(int numsel,int numset)
/* Calculation numsel from numset combinations */
{
   /* Standard combination calculation: numset!/(numsel!*(numset-numsel)!)
      However, because fact[] is a double array, we're really scared of
      rounding errors when truncating the result to a long, so we add 0.5 to
      be totally safe of avoiding any errors. */
   return((long)(fact[numset]/(fact[numsel]*fact[numset-numsel])+0.5));
}

static void init_combs()
/* Initialise factorial and combination arrays, allocate memory for the
   (drawn from upper) tickets and zero that memory to reset the combination
   counts for every possible ticket combination. */
{
   int z,y;
   double mult=1.0; /* 1! = 1 */
   long mem_size;

   /* Calculate 1! through to upper! */
   for (z=1;z<=upper;z++) { mult*=z; fact[z]=mult; }

   /* combidx[0] is referenced in comb2idx() routine and
      needs to hold zero */
   combidx[0]=0;

   /* Now calc combs(1,1) through to combs(drawn,upper)
      for faster calculations */
   for (y=0;y<=drawn;y++)
   for (z=0;z<=upper;z++)
   if (!y || !z || y>z) combs[y][z]=0; /* 0 from z, y from 0 or y>z all = 0 */
   else if (y==z) combs[y][z]=1; /* y=z means only 1 combination */
   else combs[y][z]=calc_combs(y,z); /* Everything else needs calculating */

   num_combs=combs[drawn][upper]; /* Number of ball combinations */
   /* Amount of memory to be used for tickets */
   mem_size=(num_combs+1)*2;

   /* Two bytes per ticket combination should be plenty. Allows up
      to 65,536 tickets to be bought per combination. Note that the first
      two bytes aren't used, so that we can index from 1...(drawn,upper).
      Mote that the verification algorithm doesn't use the combination array
      at all, so don't allocate the memory array if -v is supplied. */
   if (!verify)
   {
      tickets=(unsigned char *)malloc((size_t)mem_size);
      if (tickets==(unsigned char *)NULL)
      {
         (void)fprintf(stderr,"FATAL: Can't allocate %ld bytes of memory for tickets\n",mem_size);
         tidy_up();
         exit(1);
      }

      /* Zero all the ticket combination counters */
      memset((void *)tickets,0,(size_t)mem_size);
   }

#ifdef MULTITHREAD
   if (cores_used>1)
   {
      /* Work out how many combinations are present in a single "slice"
         where the slice is defined by the first number being constant
         e.g. 01 02 03 04 05 06 to 01 45 46 47 48 49 inclusive. Used when
         multi-threading to split up 14m combos into smaller job chunks. */

      save_ball_set(); /* We overwrite combidx[], so need to save it first */
      for (y=1;y<=upper-drawn+1;y++)
      {
         int tk;
         for (z=1;z<=drawn;z++) combidx[z]=y+z-1; /* Put y..y+drawn-1 in ticket */
         tk=comb2idx(); /* Find combination index number for first ticket combo */
         /* Put upper-drawn+2..upper in ticket pos 2..drawn (leave first number as y) */
         for (z=2;z<=drawn;z++) combidx[z]=upper-drawn+z;
         combo_slice[y]=comb2idx()-tk+1; /* Calculate number of combos in slice */
      }
      restore_ball_set();

      /* Fill in the rest of the array with zeroes, because there's no combos */
      for (y=upper-drawn+2;y<=upper;y++) combo_slice[y]=0;

      for (y=1;y<=drawn-2;y++)
      scaled_speed_combo[y]=(int)(
#ifdef VARIABLE_COMBO_CHUNK
      /* This will range from 66,224 Match 1 to
        4,661,272 Match 4 combos per job. Works slightly faster on 64-bit,
        so 64-bit binaries are compiled with this. */
      num_combs/speed_combo[y]
#else
      /* Fixed at 500,000 combos per job - 32-bit runs this faster than
         the variably-sized jobs */
      COMBO_CHUNK_SIZE
#endif
      );
   }
#endif /* MULTITHREAD */
}

/* Next 2 functions re-coded from:
   http://www.xtremevbtalk.com/showthread.php?t=168296
*/

#ifdef DEBUG
static void index2comb(idx)
long idx; /* Value between 1 and combs[drawn][upper] */
/* This function takes an index value into a set of combinations
   returns the array of drawn numbers in combidx[1..drawn].
   Turns out that after porting this function, it's never called
   at all! Ho hum... */
{
   int x,w=1,leave;
   long combcount=0,c;
   /* Must be within the combination size, otherwise return zeroes */
   if (idx<1 || idx>num_combs)
   { for (x=1;x<=drawn;x++) combidx[x]=0; return; }

   for (x=1;x<=drawn-1;x++)
   {
      leave=0;
      while (!leave)
      {
         c=combs[drawn-x][upper-w];
         leave=(!c || combcount+c>=idx);
         if (!leave) { combcount+=c; w++; }
      }
      combidx[x]=w++;
   }
   combidx[drawn]=w+idx-combcount-1;
}
#endif

/* This function takes a set of ascending numbers in combidx[1..drawn]
   and uses combination counts to returning a single index value between
   1 and (drawn and upper) uniquely identifying that combination. This
   index, when multiplied by 2, gives us direct access to the two chars
   used to hold the current count of tickets seen so far with that
   combination. All this effort is to save us having to have, say, a
   6-dimensional array of 49 elements in each dimension, which would use
   somewhere in the region of 26GB of RAM (i.e. (49^6)*2 bytes). Instead,
   we use only 28MB RAM, which any PC in the last 5 years is more than
   happy with. */
int comb2idx()
{
   int idx=0;
   int x,v;
   for (x=1;x<=drawn-1;x++)
   for (v=combidx[x-1]+1;v<combidx[x];v++)
   idx+=(int)(combs[drawn-x][upper-v]-1); 
   return(idx+combidx[drawn]-drawn+1);
}
   
int get_tickets_for_comb(int idx)
/* Returns the number of tickets bought for combination index */
{
   /* Yes, the return value hurts the eyes, but it's really just
      (low byte+256*high byte) of the index's count. Shifting should be
      faster than multiplying, just in case the compiler forgets to
      optimise that :-) */
   return((int)tickets[idx<<1]+(((int)tickets[(idx<<1)+1])<<8));
}

void put_tickets_for_comb(int idx,int numticks)
/* Sets the number of tickets bought for combination index */
{
   /* I used unsigned chars to play it safe here, so that bit 7 wouldn't
      get zeroed. All were doing here is splitting the new index value
      into a low and high byte and storing it in the memory array. */
   tickets[idx<<1]=(unsigned char)(numticks & 255);
   tickets[(idx<<1)+1]=(unsigned char)(numticks>>8);
}

static void out_sales_and_rollover()
/* Output the ticket sales and rollover status */
{
   (void)fprintf(stderr,"\nTicket sales = %s / ",pounds(ticket_sales));
   if (rollover)
      (void)fprintf(stderr,"rollover = %s (%s)",pounds(rollover),
                   (topup?"top-up to jackpot":"exact Super Draw jackpot"));
   else (void)fprintf(stderr,"no rollover");
   (void)fprintf(stderr,"\n\n");
}

static void get_ticket_sales_and_rollover()
{
   char buff[BUFSIZ];

   /* If you've not supplied a -t option and you're not reading from stdin
      (which will ultimately give us the ticket sales when stdin hits EOF
      once all the tickets are read in), then ask for them interactively.

      If the joker at the keyboard still types "0", then set them to
      twice the number of ticket combinations to provide at least some
      semblance of a sensible value. */

   if (!ticket_sales && !read_tickets)
   {
      if (!(ticket_sales=ask_for_value("Enter ticket sales")))
         ticket_sales=num_combs*2;
      else if (ticket_sales<drawn*upper) ticket_sales=drawn*upper;
   }

   /* If -r hasn't been supplied, ask for the rollover (or exact jackpot)
      amount */
   if (rollover<0)
   {
      (void)strcpy(buff,ask_for_string("Enter rollover top-up (prefix with = for exact Super Draw jackpot)"));

      if (buff[0])
      {
         /* A '=' prefix means set an exact Super Draw jackpot instead */
         if (buff[0]=='=') { topup=0; rollover=atol(&buff[1]); }
         else rollover=atol(buff);
      } else rollover=0;
   }
}

#ifdef DEBUG
static void test_index(ival)
long ival;
/* When porting the index2comb() and comb2idx() functions, I needed to
   make sure that they would go "both ways" (i.e. index -> combination ->
   index) and still end up with the same index at the end as the one at
   the start. It turns out I probably didn't need to bother, because I never
   call the index2comb() function at all, unless DEBUG is defined! */
{
   int l;
   (void)fprintf(stderr,"Index %ld = ",ival);
   index2comb(ival);
   for (l=1;l<=drawn;l++) (void)fprintf(stderr,"%02d ",combidx[l]);
   (void)fprintf(stderr,"\n");
   (void)fprintf(stderr,"Reverse index = %ld\n",(long)comb2idx());
}
#endif

static void usage(char *progname)
/* Tell the user what to type next time... */
{
   (void)fprintf(stderr,"%s %s - Simulates ticket sales and a live draw for the main UK Lotto\n",progname,VERSION);
   (void)fprintf(stderr,"(C) Richard K. Lloyd 2010 <webmaster@lottery.merseyworld.com>\n\n");
   (void)fprintf(stderr,"Syntax: %s [-?|-h] [-b ball1 [-b ball2..]] ",progname);
#ifdef MULTITHREAD
   (void)fprintf(stderr,"[-c cores] [-f file] ");
#endif
   (void)fprintf(stderr,"[-g] [-i]\n");
   (void)fprintf(stderr,"        [-l] [-m] [-n balls] [-O] [-o] [-p] [-r [=]rollover] [-s seed]\n");
   (void)fprintf(stderr,"        [-T 3-match-prize] [-t sales] [-u] [-v] [-z]\n\n");
   
   (void)fprintf(stderr,"-? or -h displays this syntax message.\n");
   (void)fprintf(stderr,"-b balln specifies the nth ball in drawn order (ball%d is the bonus ball).\n",totdrawn);
   (void)fprintf(stderr,"Any omitted ball will be prompted for.\n");
#ifdef MULTITHREAD
   (void)fprintf(stderr,"-c cores will use the specified number of CPU cores (1-%d, default=%d).\n",(num_cores>MAX_CPU_CORES?MAX_CPU_CORES:num_cores),cores_used);
#endif
   (void)fprintf(stderr,"-f file reads the ball numbers from the specified file instead of from the\n");
   (void)fprintf(stderr,"keyboard. The file is monitored every %.1f seconds for any changes.\n",FILE_SLEEP);
   (void)fprintf(stderr,"-g uses " ZENITY_BINARY " to display input dialogue boxes (needs X11).\n");
   (void)fprintf(stderr,"-i reads tickets from standard input (%d ascending unique integers between\n",drawn);
   (void)fprintf(stderr,"1 and %d per line). If omitted, tickets will be randomly generated instead.\n",upper);
   (void)fprintf(stderr,"-l will show losing ticket counts (i.e. those matching less than %d numbers).\n",min_match);
   (void)fprintf(stderr,"-m will omit any timings, which is mainly useful for keeping the output\n");
   (void)fprintf(stderr,"reproducible and therefore comparable in test suites. You should preferably\n");
   (void)fprintf(stderr,"also set a specific random seed via -s to keep the output consistent.\n");
   (void)fprintf(stderr,"-n balls sets the number of balls (%d-%d) in the machine, with the default = %d.\n",totdrawn+1,MAX_UPPER,upper);
   (void)fprintf(stderr,"-O outputs the prize table, totals, ticket sales and numbers to standard output\n");
   (void)fprintf(stderr,"in CSV format to be fed into another program (e.g. a spreadsheet). The\n");
   (void)fprintf(stderr,"default is to display a formatted table to stderr instead.\n");
   (void)fprintf(stderr,"-o will repeat any tickets input or generated back to standard output again.\n");
   (void)fprintf(stderr,"-p will suffix '+' or '- 'to the results table figures if more balls are yet\n");
   (void)fprintf(stderr,"to be drawn and the figure can still change (in the indicated direction).\n");
   (void)fprintf(stderr,"-r [=]rollover sets either the rollover top-up amount or, if it is prefixed\n");
   (void)fprintf(stderr,"with '=', the exact Super Draw jackpot prize pool amount.\n");
   (void)fprintf(stderr,"-s seed sets the random number generator seed to allow results to be\n");
   (void)fprintf(stderr,"easily reproduced. If omitted, the seed is set based on the current time/PID.\n");
   (void)fprintf(stderr,"-T 3-match-prize sets the 3-match prize in pounds (%d-%d, default=%d).\n",THREE_MATCH_PRIZE_MIN,THREE_MATCH_PRIZE_MAX,THREE_MATCH_PRIZE_DEFAULT);
   (void)fprintf(stderr,"-t sales sets the ticket sales. If omitted, you will be prompted for the value.\n");
   (void)fprintf(stderr,"-u omits the bonus ball from the draw entirely (not tested at all yet!).\n");
   (void)fprintf(stderr,"-v switches on alternative verification algorithm that is only run *after* the\n");
   (void)fprintf(stderr,"the draw's numbers are all input. This allows on-the-fly calculation of the\n");
   (void)fprintf(stderr,"matches for each ticket input and therefore is used to confirm the matches\n");
   (void)fprintf(stderr,"generated by the default 'live draw' (which doesn't know any draw numbers until\n");
   (void)fprintf(stderr,"after all the tickets are read in). This option also switches on -i,\n");
   (void)fprintf(stderr,"so that the tickets to be verified are fed in via standard input.\n");
   (void)fprintf(stderr,"This means that the output from running with -i (and without -v) can be fed via\n");
   (void)fprintf(stderr,"a pipe to the same program running with -v, eliminating the need for any\n");
   (void)fprintf(stderr,"potentially very large temporary ticket files.\n");
   (void)fprintf(stderr,"-z displays individual prize amounts in the results table as soon as they\n");
   (void)fprintf(stderr,"calculable, even if they will change when later balls are drawn. The default\n");
   (void)fprintf(stderr,"is to only show each prize amount once it has reached its final value.\n");
   tidy_up();
   exit(1);
}

int add_next_number(int next_ball)
/* Add the next (or remove the previous) drawn number to/from the
   combidx[] array, assuming it's in range and not a duplicate.
   Return 1 if a valid number is added, 0 otherwise */
{
   /* See if we should "delete" already drawn numbers
      (0 = do nothing, -1 = remove one, -2 = remove two etc.).
      If "too negative", just remove all drawn numbers. */
   if (next_ball<=0)
   {
      numsofar+=next_ball; if (numsofar<0) numsofar=0;
   }
   else
   /* Validate the ball is OK (e.g. between 1 and 49
       and not a duplicate) */
   if (numsofar<=drawn && next_ball<=upper)
   { 
      int bl,gotdup=0;
      for (bl=1;bl<=numsofar && !gotdup;bl++)
      gotdup=(combidx[bl]==next_ball);
      if (!gotdup)
      {
         combidx[++numsofar]=next_ball;
#ifdef MULTITHREAD
         /* If all balls supplied, child threads don't
            have to usleep() waiting for jobs, because
            they'll turn up "immediately" (i.e. without
            waiting for a human to type in any remaining
            balls not supplied with the -b option) */
         wait_for_jobs=(numsofar<=drawn);
#endif
         return(1);
      }
   }
   return(0);
}

static void get_options(int argc,char **argv)
/* Parse the qualifiers passed to the program */
{
   int param,newval;

   /* Not drawing numbers from a file yet */
   drawn_file[0]='\0'; draw_from_file=0;
   draw_file_stamp=0; draw_file_size=0;

#ifdef MULTITHREAD
    /* Include -c flag if multi-threading */
#define EXTRA_COMMAND_FLAGS "c:"
   /* Use all the CPU cores (up to MAX_CPU_CORES) by default */
   cores_used=(num_cores>MAX_CPU_CORES?MAX_CPU_CORES:num_cores);
#else
    /* No -c flag (since will use 1 CPU core) if no threads */
#define EXTRA_COMMAND_FLAGS ""
   cores_used=1; /* Single CPU core used, no multi-threading */
#endif /* MULTITHREAD */
   
   while ((param = getopt(argc,argv,"+b:f:ghilmn:Oopr:s:T:t:uvwz?" EXTRA_COMMAND_FLAGS)) != EOF)
   {
      switch (param)
      {
         case 'h':
         case '?': /* Bad option or help requested */
                   usage(argv[0]);
                   break;
         case 'b': /* Add specified number, but only if not drawing from file */
                   if (draw_from_file)
                      (void)fprintf(stderr,"-b and -f options are mutually exclusive, so -b ignored.\n");
                   else (void)add_next_number(atoi(optarg));
                   break;
#ifdef MULTITHREAD
         case 'c': /* If multi-threading, then specify the number of CPU cores
                      we should use (1-min(MAX_CPU_CORES,num cores PC has)) */
                   cores_used=atoi(optarg);
                   if (cores_used<1)
                   {
                      (void)fprintf(stderr,"Can't use less than one CPU core, so will use 1, not %d\n",cores_used);
                      cores_used=1;
                   }
                   else
                   if (cores_used>MAX_CPU_CORES || cores_used>num_cores)
                   {
                      int new_cores_used=(cores_used>MAX_CPU_CORES?MAX_CPU_CORES:cores_used);
                      if (new_cores_used>num_cores) new_cores_used=num_cores;
                      (void)fprintf(stderr,"Can use a maximum of %d CPU cores, so will use %d, not %d\n",
                                    new_cores_used,new_cores_used,cores_used);
                      cores_used=new_cores_used;
                   }
                   break;
#endif
         case 'f': /* Read drawn numbers from a monitored file, but only
                      if -b hasn't already been specified. */
                   if (numsofar)
                      (void)fprintf(stderr,"-b and -f options are mutually exclusive, so -f ignored.\n");
                   else
                   {
                      (void)strcpy(drawn_file,optarg);
                      draw_from_file=1;
                   }
                   break;
         case 'g': /* Provide a GUI, as long as zenity present and $DISPLAY set */
                   request_gui();
                   break;
         case 'i': /* Read tickets via stdin instead of random generation */
                   read_tickets=1;
                   break;
         case 'l': /* Show losing tickets (i.e. < min match for a win) */
                   show_losers=1;
                   break;
         case 'm': /* Don't show timings, achieve consistent output */
                   timings=0;
                   break;
         case 'n': /* Set number of balls in the machine between 1-MAX_UPPER */
                   newval=atoi(optarg);
                   if (newval<totdrawn+1 || newval>MAX_UPPER)
                      (void)fprintf(stderr,"Number of balls must be between %d and %d - ignored, using default of %d\n",
                                    totdrawn+1,MAX_UPPER,upper);
                   else upper=newval;
                   break;
         case 'O': /* Output final table as CSV data rather than
                      prettily formatted */
                   csv_output=1;
                   break;
         case 'o': /* Duplicate input/generated tickets to stdout as well */
                   write_tickets=1;
                   break;
         case 'p': /* Show +/- suffix on category winners where needed */
                   plusminus=1;
                   break;
         case 'r': /* Set rollover or exact jackpot prize pool (= prefix) */
                   if (optarg[0]=='=')
                   { topup=0; rollover=labs(atol(&optarg[1])); }
                   else
                   { topup=1; rollover=labs(atol(optarg)); }
                   break;
         case 's': /* Set random seed for consistent random tickets */
                   randseed=labs(atol(optarg));
                   break;
         case 'T': /* Set 3-match prize (default is 10 pounds) */
                   three_match_prize=atoi(optarg);
                   if (three_match_prize<THREE_MATCH_PRIZE_MIN ||
                       three_match_prize>THREE_MATCH_PRIZE_MAX)
                   {
                      (void)fprintf(stderr,"%s 3-match prize ignored (must be %s%d-%s%d) - set to default of %s%d\n",
                                    pounds(three_match_prize),
                                    POUND_SYMBOL,THREE_MATCH_PRIZE_MIN,
                                    POUND_SYMBOL,THREE_MATCH_PRIZE_MAX,
                                    POUND_SYMBOL,THREE_MATCH_PRIZE_DEFAULT);
                      three_match_prize=THREE_MATCH_PRIZE_DEFAULT;
                   }
                   break;
         case 't': /* Set ticket sales */
                   ticket_sales=labs(atol(optarg));
                   break;
         case 'u': /* Do not include the bonus ball in the draw at all */
                   have_bonus_ball=0; totdrawn=drawn;
                   break;
         case 'w': /* Show all ind. prizes, even if they can change later */
                   show_all_prizes=1;
                   break;
         case 'v': /* Switch to post-draw verification algorithm. Also
                      switch on reading tickets from stdin. */
                   verify=1; read_tickets=1;
                   break;
         case 'z': /* Show individual prize amounts in results table as soon as
                      they are calculable, even if they will change when later
                      balls are drawn. The default is to only show the amount
                      when it reaches its final value, to avoid accusations
                      that Camelot have "stolen" the prize money (e.g. match 4
                      prizes actually drops when the 5th and 6th balls are
                      drawn [more winners, same prize pool]). */
                   show_prize_early=1;
                   break;
      }
   }

#ifdef MULTITHREAD
   /* Verification code doesn't multi-thread, so just use a single core */
   if (verify) cores_used=1;
#endif
}

int main(int argc,char **argv)
{
#ifdef MULTITHREAD
   /* Find the number of CPU cores present */
   get_cpu_cores();
#endif

   /* Parse command-line options */
   get_options(argc,argv);
   /* Initialise factorial/combination arrays and memory for ticket combs */
   init_combs();

#ifdef DEBUG
   /* Just a bit of testing info/debugging here, but real end-user doesn't
      need to know any of this */
   (void)fprintf(stderr,"\ncomb(%d,%d)=%s possible ticket combinations\n",
                drawn,upper,mills(num_combs));
   test_index(1);
   test_index(num_combs/2);
   test_index(num_combs);
#endif

   /* Obtain ticket sales and rollover/exact figure */
   get_ticket_sales_and_rollover();

   if (read_tickets)
   {
      /* Read tickets from stdin */
      feed_tickets_from_stdin();
      /* Display the number of tickets read in and the rollover figure */
      out_sales_and_rollover();
      /* Verification algorithm, if chosen with -v, was actually run when
         the tickets were read in (because the draw's numbers are known
         beforehand), so all we do here is just display the total matches
         for each category. */
      if (verify) show_verify_table();
   }
   else
   {
      /* Display the number of tickets read in and the rollover figure */
      out_sales_and_rollover();
      /* Randomly generate the tickets */
      generate_tickets();
   }

   /* If we didn't specify -v for the verification algorithm, we use the
      'live draw' algorithm instead, where none of the draw's numbers are
      known before the tickets are read in and each of the numbers is drawn
      one at a time and the running totals in each category are shown after
      each ball. */
   if (!verify) perform_draw();

   /* Free up any memory allocated and clear up any threads left */
   tidy_up();
   return(0);
}
