/* generate.c - (C) Richard K. Lloyd 2010 <webmaster@lottery.merseyworld.com>

   Generate "random" tickets for a simulated draw or read them in from stdin.
   Skew the randomness by using the percentage column of data here:
   http://lottery.merseyworld.com/Analysis/PopularAll.html

   Allow a percentage variation from the above data as well since we need
   at least some randomness.
*/

/* Standard program definitions */
#include "draw.h"
/* This is a generated file containing a percentage array that shouldn't
   be manually edited */
#include "percent.h"

/* Variables declared in other source files */
extern int upper,drawn,write_tickets,verify,totdrawn,
           have_bonus_ball,min_match,timings,show_losers;
extern long ticket_sales,randseed;
extern int combidx[MAX_DRAWN+2],saveidx[MAX_DRAWN+2];

/* Functions declared in other source files */
extern int comb2idx(void);
extern int get_tickets_for_comb(int idx),insert_number(int done,int newnum);
extern void put_tickets_for_comb(int idx,int numticks),
            start_micro_time(void),tidy_up(void),
            save_ball_set(void),restore_ball_set(void);
extern double end_micro_time(char *tmess);
extern char *mills(long num),*pounds(long num);

/* Camelot might like to tell the public how many people bought a ticket
   containing the ball just drawn - that's what this array is for */
long num_picked[MAX_UPPER+1]; /* Number of tickets containing each number */

/* We need to scale up the tricky percentages in percent.h to something
   resembling a proportion of ticket tales */
static long scaled_percents[MAX_UPPER+1]; /* Scaled by ticket sales */

/* Number of matches at each category (i.e. 0-match through to jackpot).
   Used by the alternative verification algorithm (-v option) to accumulate
   match counts as each ticket is read in */
static int matches[MAX_DRAWN+2];

/* Number of tickets either read in from stdin or randomly generated so far */
static long ticksdone;

/* Forward declarations to shut up some compilers */
void feed_tickets_from_stdin(void);
void show_verify_table(void);
void normalise_percents();
void generate_tickets(void);

static void output_progress(char *mess)
/* We need to keep the user informed as to how ticket reading/generation
   is going in, so give them a total read in so far every 1m tickets.

   This function will also:
   - Output the current ticket to stdout if -o was supplied.
   - Increment the count for the current ticket's combination index.
   - Increment the num_picked[] frequency array elements for each of the
     ticket's numbers.
*/
{
   int tl;

   /* Output progress message for every 1m tickets processed */
   if (timings && !(ticksdone%1000000))
      (void)fprintf(stderr,"%12s tickets %s...\n",mills(ticksdone),mess);

   /* Output the current ticket if -o was supplied */
   if (write_tickets)
     for (tl=1;tl<=drawn;tl++)
     (void)printf("%02d%s",combidx[tl],(tl<drawn?" ":"\n"));

   /* Increment ticket sale for combination (not for verification algorithm
      though) */
   if (!verify)
   {
      /* Get combination index */
      tl=comb2idx();
      put_tickets_for_comb(tl,get_tickets_for_comb(tl)+1);
   }

   /* Update number of tickets containing picked numbers */
   for (tl=1;tl<=drawn;tl++) num_picked[combidx[tl]]++;
}

static void short_single_line()
/* Output a short line of minuses - used as a boundary */
{
   (void)fprintf(stderr,"--------------------\n");
}

static void reset_arrays()
/* Reset the frequency counts of the numbers that can appear on a ticket. */
{
   int r1;
   for (r1=1;r1<=upper;r1++) num_picked[r1]=0;
}

void feed_tickets_from_stdin(void)
/* Read lines of tickets from standard input, throwing away any that aren't
   6 ascending integers between 1 and upper. If -v was supplied, increment
   the appropriate matches[] array element based on the number of balls
   the matched. */
{
   /* Lots of fiddly ints declared outside loops to avoid declarations being
      repeated */
   int ml,num_mat,iv,mat,dn,vl,gd;

   /* Temporary buffer for stdin lines */
   char tbuff[BUFSIZ];

   /* Start the clock running! */
   start_micro_time();
   /* Copy combidx[] (the drawn balls so far) to saveidx[] because we're
      going to overwrite combidx[] inside the ticket scanning loop */
   save_ball_set();
   /* Reset num_picked[] array so frequency counts of numbers seen on tickets
      are all set to zero. */
   reset_arrays();
    
   /* No tickets read in so far */
   ticksdone=0;
   /* No tickets so far, so zero the categories for matches */
   for (ml=0;ml<=totdrawn;ml++) matches[ml]=0;

   /* Keep looping until stdin hits EOF (returns a null string) */
   while (fgets(tbuff,BUFSIZ,stdin)!=(char *)NULL)
   {
      /* Quick and dirty way to get an entire line of numbers into
         an array in one statement */
      (void)sscanf(tbuff,"%02d %02d %02d %02d %02d %02d",
                   &combidx[1],&combidx[2],&combidx[3],
                   &combidx[4],&combidx[5],&combidx[6]);
#ifdef DEBUG
      (void)fprintf(stderr,"Ticket = %02d %02d %02d %02d %02d %02d\n",
                    combidx[1],combidx[2],combidx[3],
                    combidx[4],combidx[5],combidx[6]);
#endif

      /* Validate numbers are OK by taking each number read in and
         inserting it into combidx[] array as if we hadn't seen it before.
         Why do this? Well, the validation code (numbers must be unique,
         must be 1..upper, need to be sorted if not in order, can't have
         more than 6 balls) is all in the insert_number() routine, so why
         not use it? */
      gd=1;
      for (vl=1;vl<=drawn && gd;vl++)
      gd=insert_number(vl-1,combidx[vl]);

      /* Only count the ticket if it passed validation */
      if (gd)
      {
         /* Yep, it's good, so that's one more read in and we
            should update various arrays and maybe report progress */
         ticksdone++; output_progress("read in");

         /* -v supplied, so run the alternative verification algorithm,
            which is pretty straightforward because we already know
            the drawn balls. */
         if (verify)
         {
            num_mat=0; /* No matches yet */
            /* Compare the 6 drawn numbers with the 6 ticket numbers */
            for (vl=1;vl<=drawn;vl++)
            {
               mat=0; dn=saveidx[vl];
               for (iv=1;iv<=drawn && !mat && dn>=combidx[iv];iv++)
               if (dn==combidx[iv]) mat=1;
               if (mat) num_mat++;
            }

            /* Now look for the bonus ball. Yes, it's similar code to
               the loop above, but just different enough to make it
               tricky to merge it into one. */
            if (have_bonus_ball && num_mat==drawn-1)
            {
               mat=0; dn=saveidx[totdrawn];
               for (iv=1;iv<=drawn && !mat && dn>=combidx[iv];iv++)
               if (dn==combidx[iv]) mat=1;
               if (mat) num_mat=totdrawn;
            }
            matches[num_mat]++; /* Increment the appropriate no. of matches */
         }
      }
   }

   /* Yes, you can supply -i and then a different -t value to the number of
      valid tickets in stdin. It's allowed, but will generate a warning. */
   if (ticket_sales>0 && ticket_sales!=ticksdone)
   {
      (void)fprintf(stderr,"WARNING: Was expecting %s tickets, but read in ",mills(ticket_sales));
      (void)fprintf(stderr,"%s\n",mills(ticksdone));
   }
   ticket_sales=ticksdone;

   /* Bring back the draw's original balls back into the combidx[] array */
   restore_ball_set();
   /* Stop the clock! */
   (void)end_micro_time("ticket reading");

   /* If you get here with no valid tickets after reading from stdin, then
      it's totally game over and the final table will just be full of
      zeroes, so don't even bother to show the tables. */
   if (!ticket_sales)
   {
      (void)fprintf(stderr,"FATAL: No tickets read in, so can't continue\n");
      tidy_up();
      exit(1);
   }
}

void show_verify_table(void)
/* Show the verification table if -v was supplied */
{
   int ml;
   long winners=0; /* No winners so far */
   /* Heading for table */
   short_single_line();
   (void)fprintf(stderr,"Verified match table\n");
   short_single_line();

   /* Show in most match order (well, OK, we actually show the Jackpot
      before the 5+Bonus winners) */
   for (ml=totdrawn;ml>=0;ml--)
   /* If -l supplied or we're on a winners line, then do this line */
   if (show_losers || ml>=min_match)
   {
   
      int zl=ml; /* The index into matches[] that we'll use for this line */
      switch (ml)
      {
         case 7: /* Jackpot is actually matches[6] */
                 (void)fprintf(stderr,"Jackpot"); zl=drawn;
                 break;
         case 6: /* If -u supplied, there's no bonus ball, so this line
                    would be the Jackpot line instead (matches[6]).
                    If we *do* have a bonus ball, then it would be
                    the 5+Bonus line (matches[7]). */
                 if (have_bonus_ball)
                 { (void)fprintf(stderr,"5+bonus"); zl=totdrawn; }
                 else (void)fprintf(stderr,"Jackpot");
                 break;
        default: /* Matches 5 down to 0 are simply matches[ml] */
                 (void)fprintf(stderr,"%d-match",ml);
                 break;
      }
      /* Display this line's matches */
      (void)fprintf(stderr,": %11s\n",mills(matches[zl]));
      /* If we're >= min_match matches, then add them to the winners total */
      if (ml>=min_match) winners+=matches[zl];
      /* If we're on the min_match boundary, then display the total number
         of winners */
      if (ml==min_match)
      {
         short_single_line();
         (void)fprintf(stderr,"Winners: %11s\n",mills(winners));
         /* We don't want a divider, unless we're showing losers (0-2
            matches) below as well. */
         if (show_losers) short_single_line();
      }
   }

   /* If -l supplied, display the number of losers */
   if (show_losers)
   {
      short_single_line();
      (void)fprintf(stderr,"Losers:  %11s\n",mills(ticket_sales-winners));
   }
   /* Display the ticket sales */
   (void)fprintf(stderr,"Sales:  %*s\n",POUND_LENGTH+11,pounds(ticket_sales));
   short_single_line();
}

void normalise_percents()
/* Convert percentages from percent.h into number of tickets that can
   contain each number (by scaling up based on ticket sales). Include a
   maximum random variation of +/- (ticket_sales/(upper*upper)) tickets for
   each number.
*/
{
   double sum_percents=0.0; /* Total of percentages */
   long ticksleft; /* Number of tickets left to allocate */
   int tl; /* Loop variable */
   long rand_var; /* +/- variation for the number of tickets where each
                     number is present */

   /* 6 numbers on a ticket, so multiply by 6 and also by a further 10% to
      ensure there's no struggle to generate the last few thousand tickets */
   ticksleft=(long)(ticket_sales*drawn*1.1);

   /* Vary the number of tickets allocated per number by this amount */
   rand_var=ticksleft/((long)upper*(long)upper);

   /* Seed the random number generator as randomly as we can manage */
   if (randseed<0)
   {
      /* Split this into 2 statements to keep icc quiet */
      randseed=time((time_t *)NULL);
      randseed^=((long)getpid()<<16);
   }
   srand48(randseed);

   /* Get the grand total of all the percents lifted from the Web page
      for the 'upper' numbers */
   for (tl=1;tl<=upper;tl++) sum_percents+=percent_wins[tl];

   /* For each number, calculate how many times it should appear in
      all the tickets that will be randomly generated */
   for (tl=1;tl<=upper;tl++)
   {
      /* base is the pivot point - number of tickets will be somewhere in
         the range (base-rand_var..base+rand_var) */
      long these_ticks=0,base=(long)(ticksleft*percent_wins[tl]/sum_percents);
      if (tl<upper)
       /* Not the last number, so try to find a suitable number of tickets
          within the range to generate */
       while (these_ticks<=0 || these_ticks>ticksleft)
       these_ticks=base-rand_var+(long)(drand48()*rand_var*2);
      else these_ticks=ticksleft; /* Last one, so allocate all that's left */
      /* Save the number of tickets decided on and deduct them from what's
         left */
      scaled_percents[tl]=these_ticks;
      ticksleft-=these_ticks; sum_percents-=percent_wins[tl];
#ifdef DEBUG
      (void)fprintf(stderr,"%2d = %8ld tickets\n",tl,these_ticks);
#endif
   }
#ifdef DEBUG
   (void)fprintf(stderr,"Ticket sales = %ld tickets (%ld left)\n",ticket_sales,ticksleft);
#endif
}

static int gen_random_num()
/* Return a random integer between 1 and upper */
{
   int limited=1,rn;
   /* Need a loop in here in case the random int we pick is "overused" */
   while (limited)
   {
      /* drand48() returns a double between 0 and 1, so multiply by upper,
         add one, truncate it to an integer and then you get a nice whole
         random number between 1 and upper. */
      rn=(int)(drand48()*upper)+1;
      /* Make sure we haven't used this number more than we wanted */
      limited=(num_picked[rn]>=scaled_percents[rn]);
   }
   return(rn);
}

void generate_tickets(void)
/* Generate the number of random tickets specified by the -t option */
{
   int x;

   /* Start timing... */
   start_micro_time();
   /* Save combidx[] so it can be restored later */
   save_ball_set();
   /* Adjust percentages into proportions of tickets sales in each
      which each number will appear */
   normalise_percents();
   /* Reset num_picked[] array (number frequencies across all tickets) */
   reset_arrays();

   /* Ticket sales loop */
   for (ticksdone=1;ticksdone<=ticket_sales;ticksdone++)
   {
      /* First number is easy - straight in there from the random number
         function */
      combidx[1]=gen_random_num();
      /* The rest of the 2..drawn numbers have to avoid clashing with the
         previous 1..drawn-1 numbers already drawn. Yes, this code is
         actually similar to that in insert_number(), but I don't want to
         use that because of its stderr messages when things aren't
         "perfect".*/
      for (x=2;x<=drawn;x++)
      {
         int n,gotnumb=0,p,s;
         /* Keep generating a random number until not of the previous x-1
            numbers drawn match it. */
         while (!gotnumb)
         {
            n=gen_random_num();
            p=x-1; while (combidx[p]>n) p--; 
            gotnumb=(combidx[p]!=n);
         }
         p++;
         /* Move larger numbers up one index to make space */
         if (p<x)
         {
            for (s=x;s>p;s--) combidx[s]=combidx[s-1];
         }
         /* Store the drawn number in the appropriate slot - this keeps
            everything in ascending order after the above shift. */
         combidx[p]=n;
      } 

      /* Tell the user what's going on occasionally and also update
         various counts */
      output_progress("randomly generated");
   }
   (void)fprintf(stderr,"\n");
   /* Restore combidx[] */
   restore_ball_set();
   /* Stop timing... */
   (void)end_micro_time("ticket generation");
}
