/* live.c - (C) Richard K. Lloyd 2010 <webmaster@lottery.merseyworld.com>

   Perform live draw by inputting each number drawn and generating "instant"
   stats */

/* Common include file */
#include "draw.h"

/* Variables declared in other source files */
extern int drawn,upper,min_match,numsofar,have_bonus_ball,totdrawn,
           timings,show_losers,plusminus,cores_used,show_all_prizes,
           csv_output,three_match_prize;
extern int combidx[MAX_DRAWN+2];
extern long num_picked[MAX_UPPER+1],combs[MAX_UPPER+1][MAX_UPPER+1];
extern long rollover,ticket_sales,topup;

/* Functions declared in other source files */
extern int get_tickets_for_comb(int idx);
extern int comb2idx(void);
extern int add_next_number(int next_ball);
extern long ask_for_value(char *prompt);

#ifdef MULTITHREAD
/* Multi-threaded code needs some variables and functions from pthread.c */
extern void spawn_threads(int balls_sofar),kill_all_children();
extern void *thread_handler(void *threadptr);
extern void put_jobs_on_queue(int num_matches,int faked_ball);
extern void batch_fake_jobs(int fire_up_threads),stop_fake_jobs(int latestnum);
extern int job_status[MAX_JOBS+1],job_matches[MAX_JOBS+1],
           fake_ball[MAX_JOBS+1],job_thread[MAX_JOBS+1],
           start_combo_num[MAX_JOBS+1],end_combo_num[MAX_JOBS+1],
           match_counts[MAX_JOBS+1],combo_slice[MAX_UPPER+1];
extern int numjobs,more_jobs_to_come,wait_for_jobs,all_jobs_done;
extern int *thread_running;

#ifdef VARIABLE_COMBO_CHUNK
/* This array is the length of time match 1..4 combinations take to compute
   relative to each other when 7 balls are drawn from 49 and a single
   CPU core is used. */
int speed_combo[MAX_DRAWN-1]={0,211,107,25,3};
#endif

/* Scaled version of speed combo array, with the chunks adjusted based on
   how fast they are at n = matches 1..4 i.e. num_combs/speed_combo[n] */
int scaled_speed_combo[MAX_DRAWN-1];

/* Time limit for parent process inside thread_handler() function */
double parent_time_limit;

#endif /* MULTITHREAD */

/* Save combidx[] in saveidx[] so that the former can be overwritten
   during calculations and then restored from the later afterwards */
int saveidx[MAX_DRAWN+2];

/* Do the match checking in a slightly strange order:

   Match 2-4 Children: Lots of jobs to keep them busy initially
   5+bonus     Parent: Match 5's need to know 5+bonus winners, so done 1st
   Match 5     Parent: Done as soon as we know 5+bonus winners
   Match 6     Parent: Single combo match, so nearly "instant"
   Match 1   Children: Top-up with more jobs so they don't run out
               Parent: Also joins in processing "child" jobs
   Match 0     Parent: Ticket sales minus matches 1-6, so "instant" and
                       has to be last as you'd expect
*/
#define PARENT_JOINS 1 /* At which match level a parent becomes a "child" */
static int match_order[MAX_DRAWN+1]={2,3,4,7,5,6,PARENT_JOINS,0};

/* Record total number of winners between 3-match and 6-match tiers */
int total_winners=0;

/* Start clock time */
static double start_lmt;

/* Used for live draw timing calculations */
static double min_time=1E9,max_time=0.0,tot_time=0.0;
static int min_time_ball=0,max_time_ball=0;

/* Category winners and prizes arrays */
static long catwin[MAX_DRAWN+2],przwin[MAX_DRAWN+2],totcatwin,totprzwin,
            saved_catwin[MAX_DRAWN+2];

#ifdef DEBUG
/* Grand total of combinations scanned for all lines of the table */
int grandscan;
#endif

int draw_from_file; /* Are we monitoring a file for drawn numbers? */
char drawn_file[BUFSIZ]; /* File to monitor via -f option */
time_t draw_file_stamp; /* Timestamp of monitored file */
long draw_file_size; /* Size (in bytes) of monitored file */

/* Keep track of highest 1st number slice for used for latest number
   by faked jobs (for matches 1-4) */
int high_comb[MAX_DRAWN+1];

/* Keep running total of cat matches for latest number
   worked out by faked jobs (for matches 1-4) */
int fake_comb[MAX_DRAWN+1];

/* Forward declarations that compilers like to see */
double micro_time(void);
void start_micro_time();
double end_micro_time(char *tmess);
char *mills(long num);
char *pounds(long num);
int insert_number(int done,int newnum);
void save_ball_set(void);
void restore_ball_set();
void display_calcs(int sofar);
void perform_draw(void);

static void single_line()
/* Display a long single line of minuses */
{
   (void)fprintf(stderr,"------------------------------------------------\n");
}

static void double_line()
/* Display a long single line of equals */
{
   (void)fprintf(stderr,"================================================\n");
}

double micro_time(void)
/* Useful routine to return current time in double precision seconds since
   1st Jan 1970 down to an accuracy of one microsecond (OK, admittedly,
   the system clock may not actually jump by that unit!). */
{
   struct timeval micro_now;
   if (gettimeofday(&micro_now,(struct timezone *)NULL))
     /* Fall back to whole seconds if gettimeofday() fails */
     return((double)time((time_t *)NULL));
   else
     return((double)((double)micro_now.tv_sec+(double)micro_now.tv_usec/1000000.0));
}

void start_micro_time()
/* Start the (very accurate) clock by recording the start time to tiny
   fractions of a second */
{
   start_lmt=micro_time();
}

double end_micro_time(char *tmess)
/* Stop the (very accurate clock) by recording the end time to tiny
   fractions of a second and then subtracting the start time from it,
   displaying the result in seconds to 6 decimal places. Allows us to see
   if we've made even tiny improvements to the speed of the code. */
{
   double tot_lmt=micro_time()-start_lmt;
   if (timings) (void)fprintf(stderr,"Time for %s = %.06f seconds\n\n",tmess,tot_lmt);
   return(tot_lmt);
}

static char *disp_millions(double num,int off)
/* Display the passed double number as a comma-separated (if >=1000) number.
   I use this on my lottery site, but I've long since forgotten much of
   what it does, so the description ends here... */
{
   static char mls[32];
   int p=0,l;
   int mil;
   if (off==-3) num+=0.005; else num+=0.5;
   mil=(int)(num/1000000);
   if (mil>=1000)
   (void)sprintf(mls,"%03d,%03d,%03d,%03d",
           (int)((mil+0.5)/1000),
           mil%1000,
           (int)((num-(double)mil*1000000.0)/1000),
           (int)(num-(double)((int)(num/1000))*1000.0));
   else
   (void)sprintf(mls,"%03d,%03d,%03d",
           mil,
           (int)((num-(double)mil*1000000.0)/1000),
           ((int)(num))%1000);
   if (off==-3)
   {
      char tbuf[8];
      while (num>MAXINT) num-=MAXINT;
      (void)sprintf(tbuf,".%02d",(int)((num-(int)num)*100));
      (void)strcat(mls,tbuf);
   }
   l=(int)strlen(mls)-1;
   while (p<l && (mls[p]=='0' || mls[p]==',') && mls[p+1]!='.')
   {
      mls[p++]=' ';
   }
   if (off<0) off=p;
   return(&mls[off]);
}

char *mills(long num)
/* Display the passed long number as a comma-separated (if >=1000) number. */
{
   return(disp_millions((double)num,-2));
}

char *pounds(long num)
/* Display the passed long number as a comma-separated (if >=1000) number,
   prefixing it with a pound symbol as well. */
{
   static char pnds[32];
   (void)sprintf(pnds,"%s%s",POUND_SYMBOL,mills(num));
   return(pnds);
}

int insert_number(int done,int newnum)
/* Assuming it's valid (i.e. is a unique number between 1 and upper) and
   we haven't already inserted totdrawn numbers, then insert the number newnum
   into the combidx[] array at position done+1, shifting any higher numbers
   up one element if necessary, thus keeping the array in ascending order.
   Returns 0 if the operation failed and != 0 if it succeeded. */
{
   if (done>totdrawn-1)
     (void)fprintf(stderr,"Can't insert any more numbers (done %d already, so rejected %d)\n",done,newnum);
   else
   if (newnum<1 || newnum>upper)
     (void)fprintf(stderr,"Number must be between 1 and %d - rejected %d\n",upper,newnum);
   else
   {
      /* Look for duplicates in any of the previous 'done' numbers */
      int p=done;
      while (p>0 && combidx[p]>newnum) p--;
      if (combidx[p]==newnum)
         (void)fprintf(stderr,"Duplicate number %d rejected\n",newnum);
      else
      {
         int s;
         /* If it's not the bonus number, see if we need to shift existing
            higher numbers up one element */
         if (done<drawn)
         {
            p++;
            /* Shift up higher numbers one element */
            for (s=done+1;s>p;s--) combidx[s]=combidx[s-1];
         } else p=done+1; /* Bonus ball, so slap it in */
         /* Not in the array yet, so insert it the appropriate place */
         combidx[p]=newnum;
         return(p);
      }
   }
   return(0);
}

void save_ball_set(void)
/* Save the combidx[] array in saveidx[], so it can be restored later on */
{
   int pl;
   /* Save original ball set for restoring later on */
   for (pl=0;pl<=totdrawn;pl++) saveidx[pl]=combidx[pl];
}

void restore_ball_set()
/* Restore the combidx[] array by copying it back from saveidx[] */
{
   int pl;
   /* Restore original ball set */
   for (pl=0;pl<=totdrawn;pl++) combidx[pl]=saveidx[pl];
}

int do_combo_matches(int thread_number,int sofar,int sl,
                     int inserted_ball,int start_point,int end_point)
/* Match 1-4's have no easy shortcut - they have to be done with combination
   scanning. We do throw in several optimisations, but the biggest speedup
   comes from splitting up the 14m combinations into "chunks" where the
   first number ranges from start_point...end_point rather than
   1..upper-drawn+1. The return value is the number tickets of bought with 'sl'
   matches from the 'sofar' balls drawn so far.

   This is the "grunt" routine for the child threads in case you're wondering,
   so it has to be thread-safe. It can use global variables if they're not
   written to by the parent process or other threads at the same time.
   Fortunately, saveidx[] - the balls drawn so far - is written to once by
   the parent process (via save_ball_set()) before any threads read it, so
   we can safely use that.
*/
{
   /* Thread-safe array of combination indexes */
   int thread_combidx[MAX_DRAWN+2],thread_saveidx[MAX_DRAWN+2];

   /* hball is the number of balls drawn so far excluding any bonus ball,
      but including any "fake" ball insertion.
      hcomb is the line number of the table (up to a maximum of 'drawn'). */
   int hball=sofar-(sofar>drawn)+(inserted_ball>0),hcomb=sl-(sl>drawn);

   /* nw holds the number of winners for this table line.
      ptr indicates the next thread_combidx[] index (1...drawn) to be
      incremented once this combination is dealt with. */
   int nw=0,pl,ptr=drawn,low,high,bl,x,v;
   int idx,tspos=0,done_insert=0;
#ifdef DEBUG
   /* No combinations scanned for this table line so far */
   int scanned=0;
#endif

   /* Insert inserted_ball (if it's >0) into correct ascending order position
      in thread_saveidx[] array. */
   if (inserted_ball)
   {
      for (pl=1;pl<=hball-1;pl++)
      {
         if (!done_insert && inserted_ball<saveidx[pl])
         {
            thread_saveidx[++tspos]=inserted_ball; done_insert=1;
         }
         thread_saveidx[++tspos]=saveidx[pl];
      }
      if (!done_insert) thread_saveidx[++tspos]=inserted_ball;
   }
   else
   for (pl=1;pl<=hball;pl++) thread_saveidx[pl]=saveidx[pl];

   /* low is the lowest value thread_combidx[drawn] can be to still achieve
      the number of matches required.
      high is similarly the highest value that thread_combidx[1] can be to
      achieve the matches target. */
   low=thread_saveidx[1]+hcomb-1; high=thread_saveidx[hball]-hcomb+1;

   /* Optimisation - skip entire combo scanning loop if there's no
      way to achieve the matches we need. */
   if (start_point<=high)
   {
      int stayinloop;

      /* Slot start_point..start_point+drawn-1 into thread_combidx[] array */
      thread_combidx[0]=0;
      for (pl=1;pl<=drawn;pl++) thread_combidx[pl]=start_point+pl-1;

#ifdef MULTITHREAD
      if (cores_used>1 && (!numjobs || thread_running[thread_number]<=0))
         stayinloop=0; else
#endif
      stayinloop=(thread_combidx[1]<=end_point);

      /* Main combination loop, only exit when the next index to
         increment is 0 */
      while (stayinloop)
      {
         /* No matches yet and not done too many either */
         int in,match=0,toomany=0;

         /* Optimisation - no matches if the last number of the combo
            is less than the lowest number possible to achieve matches or
            the first number of the combo is greater than the highest
            number possible to achieve matches. */
         if (thread_combidx[drawn]>=low && thread_combidx[1]<=high)
         {
#ifdef DEBUG
            /* Yep, we've found a combination to scan at long last */
            scanned++;
#endif
            /* Look at each number drawn so far and see if the
               current combination matches that number. Exit early
               if we match too many (i.e. more than sl). */
            for (pl=1;pl<=hball && !toomany;pl++) 
            {
               bl=thread_saveidx[pl];
               /* Leave inner scan early if either match too many
                  numbers or the combo number is greater than the
                  ball number. */
               for (in=1;in<=drawn && !toomany && thread_combidx[in]<=bl;in++)
               if (thread_combidx[in]==bl)
               {
                  /* Optimisation - flag when we have too many matches */
                  if (match==sl) toomany=in; else
                  match++; /* We have a match and it's not too many */
               }
            }
         }

         if (toomany) ptr=toomany; /* Increment from overflow match */
         else
         {
            int mdiff=sl-match; /* How many matches we fell short by */
            if (mdiff)
               /* Optimisation - 1-4 matches short, so increment the
                  (drawn+1-fallen_short)th number */
               ptr=drawn+1-mdiff;
            else
            {
               /* Exact match, so success. Add the number of
                  tickets sold for this combination. */
   
               /* Yes, this is a repeat of the comb2idx() function, but
                  with thread_combidx[] instead of combidx[]. Not easy to
                  macroise, but it's not the end of the world. */
               idx=0;
               for (x=1;x<=drawn-1;x++)
               for (v=thread_combidx[x-1]+1;v<thread_combidx[x];v++)
               idx+=(int)(combs[drawn-x][upper-v]-1);
               nw+=get_tickets_for_comb(idx+thread_combidx[drawn]-drawn+1);
               ptr=drawn; /* Increment the last number */
            }
         }
   
         /* Complicated, but basically backs the pointer up until
            it can be incremented safely. For example, thread_combidx[6]
            can't go higher than 49, thread_combidx[5] higher than 48 and
            so on. */
         while (ptr && ++thread_combidx[ptr]>upper-(drawn-ptr)) ptr--;
         if (ptr>0 && ptr<drawn)
         {
            int lp;
            /* We've backed the pointer back at least one element and
               haven't finished our combinations yet, so fill in
               thread_combidx[pointer+1]..thread_combidx[drawn] with the values
               thread_combidx[pointer]+1..thread_combindx[pointer]+drawn-pointer */
            for (lp=drawn;lp>ptr;lp--)
            thread_combidx[lp]=thread_combidx[ptr]+lp-ptr;
         }

#ifdef MULTITHREAD
         if (cores_used>1 && (!numjobs || thread_running[thread_number]<=0))
            stayinloop=0; else
#endif
         stayinloop=(thread_combidx[1]<=end_point);
      }
   }

#ifdef DEBUG
   /* Display number of combos scanned for this table line */
   (void)fprintf(stderr,"combos[%d/%d] scanned: %11s\n",sl,hball,
                 mills(scanned));
   grandscan+=scanned; /* Update grand total of scanned combos */
#endif

#ifdef DEBUG
#ifdef MULTITHREAD
   if (cores_used>1 && (!numjobs || thread_running[thread_number]<=0))
      (void)fprintf(stderr,"Aborted: do_combo_matches(thread_number=%d,hball=%d,sl=%d,inserted_ball=%d,start=%d,end=%d)=%d\n",
                    thread_number,hball,sl,inserted_ball,start_point,end_point,nw);
#endif
#endif
 
   return(nw);
}

void show_csv_results(int sofar)
/* Send results so far as CSV data to stdout (usually for piping into another program).
   Format is <field number>,<field_data>

   <field_number> values dictate the <field_data> as follows:

   0-7 = <number_of_tickets_with_field_number_matches>,<prize_amount_for_field_number>,
         <prize_pool_for_field_number>
         Note that <field_number>=6 means 5+bonus and =7 means jackpot (6-match)

     8 = <total_prize_winners>,<total_prize_pool>,<ticket_sales>

     9 = The numbers drawn so far

   Note that there's no "plus or minus" suffixes because they'll just upset any
   program or spreadsheet that reads then in. */
{
   int sl,rl;
   long totcat,totprz,przpool;

   totcatwin=0; totprzwin=0;
   for (sl=0;sl<=totdrawn;sl++)
   if (show_losers || sl>=min_match)
   {
      if (sofar>=drawn && sl>=drawn && have_bonus_ball) rl=drawn+(totdrawn-sl); else rl=sl;
      totcat=catwin[rl]; totprz=przwin[rl];
      (void)printf("%d,%ld,%ld,%ld\n",sl,totcat,totprz,(przpool=totcat*totprz));
      if (sl>=min_match) { totcatwin+=totcat; totprzwin+=przpool; }
   }
   (void)printf("%d,%ld,%ld,%ld\n%d,",
                totdrawn+1,totcatwin,totprzwin,ticket_sales,
                totdrawn+2);
   for (sl=1;sl<=sofar;sl++)
   {
      (void)printf("%02d",combidx[sl]);
      if (sl<sofar) (void)printf(",");
   }
   (void)printf("\n");
}

void show_formatted_results(int sofar)
/* Finally, the end user gets to see something as we output the
   live draw prize/winners formatted table after each ball is drawn. */
{
   int sl;

   /* Heading first */
   single_line();
   (void)fprintf(stderr," Category       Prize     Winners    Prize Pool\n");
   single_line();

   for (sl=totdrawn;sl>=0;sl--)
   /* Show the table line if either we supplied -l (to show "losers") or
      it's a winners line (i.e. at least 3-match */
   if (show_losers || sl>=min_match)
   {
      /* OK, confession time, this block of code is the messiest of the lot -
         almost trial'n'error hacked around until I got it to behave nicely.
         So apologies in advance if you can't understand things for a
         while... */
      /* Set the "real" index for winners/prizes - needs to be fudged for
         5+bonus vs. jackpot (i.e. swapped over) */
      int rl=sl;
      /* Suffix chars used to indicate if values are going to stay the same
         (a space), increase (a plus) or decrease (a minus) when the ball after
         this one is drawn. Note that s space will be used unless -p was
         supplied to turn on this "plusminus" suffix. */
      char nc=(char)(sl<min_match || sofar>drawn?' ':'-'),
           dfc=(char)((sl==drawn-1 && sofar<=drawn && sofar!=sl)?'-':'+');

      if ((sl==min_match && sofar>=min_match) ||
          (sofar>=drawn && (sl>drawn || sl==drawn-2))) nc=' ';
      if (show_all_prizes)
      {
         if (sofar==drawn && sl==drawn-1) nc='+';
      }
      if (sofar==totdrawn) nc=' ';

      if (!plusminus || (sofar>=drawn && sl!=drawn-1)) dfc=' ';
      /* Flip over 5+bonus vs. jackpot index if necessary */
      if (sofar>=drawn && sl>=drawn && have_bonus_ball) rl=drawn+(totdrawn-sl);

      /* Display the category name */
      if (sl==min_match-1 && sofar==min_match) single_line(); /* Split win v loss */
      if (sl==totdrawn) (void)fprintf(stderr," Jackpot: ");
      else if (sl<drawn) (void)fprintf(stderr," %d-match: ",sl);
      else (void)fprintf(stderr," 5+bonus: ");

      /* Just display either 3 dashes (unknown winners row) or 3 spaces
         (unknown losers row) if we can't do any calcs at all for the row yet */
      if (sofar<=drawn &&
          (sofar<rl || (sofar==rl && rl>drawn)))
         (void)fprintf(stderr,"          %c           %c             %c",nc,nc,nc);
      else
      {
         /* Set up some winner/prizes/pool variables for easy access */
         long thiswin=catwin[rl],thisprz=przwin[rl];
         long thispool=thiswin*thisprz;
 
         /* Only display the individual prize if it's 3-match (and we've drawn
            >= 3 balls) or it's 4-match or 6-match (and >= 6 balls drawn) or
            all 7 balls drawn. It's better not to display an individual prize
            amount until it can't change with any of the later balls are drawn
            (cue complaints from idiots who claim they've been "diddled" out of
            prize money). */
         if ((thisprz>0 || rl>=min_match) &&
             (sofar>=rl || rl>drawn) &&
             (sofar>=drawn || rl==min_match || show_all_prizes) &&
             (sofar!=drawn || rl!=drawn-1 || show_all_prizes))
            (void)fprintf(stderr,"%*s%c",10+POUND_LENGTH,pounds(thisprz),
                          (plusminus?nc:' '));
         else (void)fprintf(stderr,"          %c ",nc);

         /* Display number of winners so far, possibly with a plus/minus suffix
            if it can change. Note that the number of 5-match winners can drop
            when the bonus ball is drawn (because the 5+bonus winners are taken
            from the 5-match group of winners then), so that actually gets a
            minus suffix after the 5th and 6th balls are drawn. */
         (void)fprintf(stderr,"%11s",mills(thiswin));
         totcatwin+=thiswin; /* Update number of winners */
         /* Save number of winners at this point, so that match 0-2 aren't
            included */
         if (sl==min_match) total_winners=totcatwin;
         if (plusminus && (sofar<drawn || (sofar==drawn && rl>0 && have_bonus_ball)))
            (void)fprintf(stderr,"%c",(rl>0?dfc:'-'));
         else (void)fprintf(stderr," ");

         /* No prize pool for losing tickets */
         if (sl<min_match)
             (void)fprintf(stderr,"             ");
         else
         /* Tickets have won, but only display the total prize pool for
            3-match tickets (since it's 10 pounds * number of 3-matches),
            4-match and 6-match pools after the 6th or 7th ball is drawn and
            5-match/5+bonus pools only after the bonus ball is drawn */
         if ((sofar>drawn || (sofar==drawn && sl!=drawn) || rl==min_match
              || show_all_prizes) &&
             (sofar!=drawn || rl!=drawn-1 || show_all_prizes))
         {
            (void)fprintf(stderr,"%*s%c",12+POUND_LENGTH,pounds(thispool),
                          (plusminus && rl==min_match && sofar<drawn)?'+':' ');
            totprzwin+=thispool; /* Update total prize pool */
         }
         else (void)fprintf(stderr,"            %c",nc);
      }
      (void)fprintf(stderr,"\n");
      /* We want to sub-total the number of winners and prize pools once we've
         done the 3-match figures. */
      if (sofar>min_match && sl==min_match)
      {
         char chgcat=(char)((plusminus && sofar<drawn)?'+':' '),
              chgprz=(char)((plusminus && sofar<totdrawn)?'+':' ');
         double_line();
         (void)fprintf(stderr," Totals: %24s%c",mills(totcatwin),chgcat);
         (void)fprintf(stderr,"%*s%c\n",12+POUND_LENGTH,
                       pounds(totprzwin),chgprz);
         double_line();
      }
   }

   /* Display ticket sales (something Camelot refuse to do on their Web site
      and won't even tell you on the public phone line either!). */
   if (show_losers || sofar<=min_match) single_line();
   (void)fprintf(stderr," Ticket sales: %*s\n",
                 17+POUND_LENGTH,pounds(ticket_sales));
   single_line();
   (void)fprintf(stderr,"\n");
}

void display_calcs(int sofar)
/* Display live draw table after the 'sofar'th ball has been drawn.
   Yes, this function is too large... */
{
   /* Total category/prize winners and number of tickets left to match
      (set to the ticket sales initially as you'd expect) */

   /* Loop variable for each line of the table */
   int sl,mo;

   /* Is overflow rule in effect? */
   int overflow=0;

   /* Prize pool for a particular table line */
   double z=0.0;

   /* Start timing live draw for this ball */
   start_micro_time();

#ifdef DEBUG
   grandscan=0; /* Nothing scanned at all yet */
#endif

   /* No category winners/prizes yet, unless we've faked this
      already */
   for (sl=0;sl<=totdrawn;sl++)
   {
      if (sofar>drawn && saved_catwin[sl]>=0) catwin[sl]=saved_catwin[sl];
      else catwin[sl]=fake_comb[sl];
#ifdef DEBUG
      (void)fprintf(stderr,"Initial catwin[%d]=%8ld, fake_comb[%d]=%8d, high_comb[%d]=%2d\n",sl,catwin[sl],sl,fake_comb[sl],sl,high_comb[sl]);
#endif
      przwin[sl]=0;
   }

   /* 3-match prize is usually 10 pounds, but you can adjust it with -T */
   przwin[min_match]=three_match_prize;
   totcatwin=0; totprzwin=0;

   /* Only need to save the ball set once to saveidx[] */
   save_ball_set();

   /* Need to calculate the number of winners in each category so far */
   for (mo=0;mo<=totdrawn;mo++)
   {
      sl=match_order[mo];
      /* Only do something if we've drawn enough balls for this table line */
      if (sofar>=sl)
      {
         /* nw holds the number of winners for this table line.
            insbonus allows for bonus number insertion in the 5+bonus check. */
         int nw=0,pl,insbonus,threaded=0;

         switch (sl)
         {
         case 0:
           /* Match 0's are just ticket sales-(anyone who matches>0) */
           nw=ticket_sales;
           for (pl=1;pl<=totdrawn;pl++) nw-=catwin[pl];
           break;

         case 1:
         case 2:
         case 3:
         case 4:
#ifdef MULTITHREAD
           if (cores_used>1)
           {
              threaded=1;
              /* Only put jobs on queue if not doing the bonus ball */
              if (sofar<=drawn) put_jobs_on_queue(sl,0);

              if (sl==PARENT_JOINS)
              {
                 /* It's the last of the threaded rows, so let parent process
                    join in the fun and "pretend" to be a thread processing
                    jobs. Only difference is that it actually returns from the
                    thread_handler() routine, unlike child threads. */
                 PTR_CAST_TYPE parent_thread_num=0;
#ifdef DEBUG
                 (void)fprintf(stderr,"%d real jobs allocated in total\n",numjobs);
#endif
                 /* We're not putting any more jobs on the queue */
                 more_jobs_to_come=0;

                 /* Parent becomes child and runs a bunch of jobs. It'll
                    return when there's no more spare jobs left to run. */
                 if (numjobs && !all_jobs_done)
                 {
                    parent_time_limit=-1.0; /* No time limit for parent */
                    (void)thread_handler((void *)parent_thread_num);
                 }

                 /* We now have 'numjobs' potentially running in the child
                    threads. We need to know how many have completed, because
                    when 'numjobs' jobs are done, we can continue beyond this
                    point. */
                 pl=0;
                 while (pl<numjobs)
                 {
                    /* Count how many jobs are completed (we want all done) */
                    int mj;
                    pl=0;
                    for (mj=1;mj<=numjobs;mj++)
                    pl+=(job_status[mj]==IDLE_JOB_COMPLETED);

                    /* Could put a usleep(THREAD_SLEEP) in here if pl<numjobs,
                       but we're right at the end of the job queue and we
                       really need to finish right away, so yes there's
                       a few hundredths of a second of 100% CPU core hammering
                       by the parent CPU here. */
                 }

                 /* All done, so update the table rows that the threads
                    deal with */
                 all_jobs_done=1;
                 thread_running[0]=1; /* Parent still alive */
                 for (pl=1;pl<=numjobs;pl++)
                 if (job_status[pl]==IDLE_JOB_COMPLETED)
                 {
                    /* Add number of combos matched by job */
                    catwin[job_matches[pl]]+=match_counts[pl];
#ifdef DEBUG
                    (void)fprintf(stderr,"Real job #%d: match %d, slice %d-%d, w%d_wins=%d, win_tot=%ld, fg_wins=%d\n",
                                  pl,job_matches[pl],start_combo_num[pl],end_combo_num[pl],
                                  job_thread[pl],match_counts[pl],
                                  catwin[job_matches[pl]],
                                  do_combo_matches(0,sofar,job_matches[pl],
                                                   0,start_combo_num[pl],
                                                   end_combo_num[pl]));
                                  
#endif
                 }

                 /* Make sure all the children are dead */
                 thread_running[0]=0; /* Parent "dead" now */
                 kill_all_children();
              }
           }
           else
#endif /* MULTITHREAD */
#ifdef DEBUG
           {
              double begin_time=micro_time();
#endif
              /* No multi-threading or only 1 core used, so do the match 1-4
                 goodies in the parent thread */
              nw=do_combo_matches(0,sofar,sl,0,1,upper-drawn+1);
#ifdef DEBUG
              (void)fprintf(stderr,"do_combo_matches(0,%d,%d,0,1,%d) = %.06f seconds\n",
                            sofar,sl,upper-drawn+1,micro_time()-begin_time);
           }
#endif

           break;

         case 5:
           /* Match 5's are processed as follows:
              1. If (sofar==5):
                 Just compare the 5 numbers drawn so far
                 against 44 combinations involving the remaining numbers
                 that could be drawn as a 6th number. Sum the tickets bought
                 for those combinations.
              2. If (sofar==6):
                 Create 43*6 combinations by substituting each main number
                 in turn with each of the 43 numbers not drawn so far.
                 Sum the tickets bought for those combinations.
              3. If (sofar==7):
                 Repeat step 2, but then subtract the number of 5+bonus
                 winners.
           */

           if (sofar==drawn-1)
           {
              /* Drawn 5 balls so far, so simple loop to find (49-5) other
                 numbers that would be drawn as the 6th number. Insert each
                 of those numbers in turn into the ticket and add uo the
                 number of tickets sold for those 44 combinations. */
                 
              for (pl=1;pl<=upper;pl++)
              {
                 int in,oldb=0;
                 /* Need to reset combidx[] with the original drawn combo
                    (aka saveidx[]) *except* ignore the combo if a saveidx[]
                    number matches the outer scan loop (which goes from
                    1..upper) */
                 for (in=1;in<=sofar && !oldb;in++)
                 if (saveidx[in]==pl) oldb=1; else combidx[in]=saveidx[in];
                 if (!oldb)
                 {
                    /* The outer loop (pl) didn't match any of the 5 numbers
                       in the original drawn combo of 5 numbers, so insert
                       pl as a 6th number and then add on the number of
                       ticket combos sold */
                    (void)insert_number(sofar,pl);
                    nw+=get_tickets_for_comb(comb2idx());
                 }
              }
           }
           else
           {
              /* Drawn 6 main balls or 6 main balls and the bonus number so
                 far and want to know the number of match 5's we have.
                 For each ticket position 1..drawn, loop for the (49-6)
                 "other" main numbers not drawn yet and put them in each of
                 the 1..drawn position (totalling 43*6 combo checks). */
              int nm;
              /* 1..drawn original balls loop */
              for (nm=1;nm<=drawn;nm++)
              {
                 /* Replacement loop 1..upper, minus any number that appeared
                    in the 1..drawn original balls */
                 for (pl=1;pl<=upper;pl++)
                 {
                    int in,oldb=0;
                    for (in=1;in<=drawn && !oldb;in++)
                    if (saveidx[in]==pl) oldb=1; else combidx[in]=saveidx[in];
                    if (!oldb)
                    {
                       /* Replacement number is not one in the original 'drawn'
                          numbers. Move combidx[outer_loop+1]..combidx[drawn]
                          down one element to
                          combidx[outer_loop]...combidx[drawn-1] and then
                          insert replacement number at combidx[drawn], sort
                          and add the number of combos found for the ticket */
                       for (in=nm;in<drawn;in++)
                       combidx[in]=combidx[in+1];
                       (void)insert_number(drawn-1,pl);
                       nw+=get_tickets_for_comb(comb2idx());
                    }
                 }
              }
              /* Special case - if we've drawn the bonus ball as well,
                 deduct the number of 5+bonus winners from the 5-match total */
              if (sofar>drawn) nw-=(int)catwin[sofar];
           }

           break;

         case 6:
           /* Match 6's are simply the number of ticket combos bought
              as indicated by our tickets memory array */

           /* Bring back combidx[] ball set */
           restore_ball_set();
           nw=get_tickets_for_comb(comb2idx());
           break;

         case 7:
           /* 5+bonus matches just require us to substitute the bonus ball in
              for each of the 6 main numbers in turn and sum the number
              of tickets found */

           insbonus=saveidx[totdrawn]; /* Bonus ball should be inserted */

           /* Loop for each main ball position */
           for (pl=1;pl<=drawn;pl++)
           {
              int in,cnt=1;
              /* Replacement the first number with the original bonus ball */
              (void)insert_number(0,insbonus);
              /* Put back all the remaining numbers except for the outer_loop's
                 ball position */
              for (in=1;in<=drawn;in++)
              if (pl!=in) (void)insert_number(cnt++,saveidx[in]);
              /* Add the number of tickets for the 5+bonus combo we just
                 created */
              nw+=get_tickets_for_comb(comb2idx());
           }
           break;
         }

         /* Save number of matches in this table line's category (non-threaded) */
         if (!threaded) catwin[sl]=nw;
      }
   }

   /* If 6th main ball drawn, save match 1-4 category figures because
      the bonus ball won't change them. */
   if (sofar==drawn)
     for (mo=0;mo<=totdrawn;mo++)
     saved_catwin[mo]=(mo>=1 && mo<=4?catwin[mo]:0);

#ifdef DEBUG
   (void)fprintf(stderr,"Grand total scanned: %11s\n",mills(grandscan));
#endif

   /* Compute total prize pool for variable categories. Yes, it's rounded
      down to the nearest 10p, despite the fact that prizes are awarded in
      whole pounds. Go figure. */
   z=(double)((int)(10*(ticket_sales*0.45-przwin[min_match]*catwin[min_match])))/10.0;

   /* Is overflow is in effect? See Paragraph E(4) for rule on this:
      http://www.national-lottery.co.uk/player/information.do?info=playlottorules */
   if (three_match_prize*total_winners>ticket_sales*0.45) overflow=(int)(z/total_winners);
   else overflow=0;

   /* Now compute the prize amounts as best as we can */
   for (sl=totdrawn;sl>=0;sl--)
   {
      long roll=0,exact=0;
      int p=0,d;
      /* Retrieve the number of matches for this table line into a temp
         variable. Because one of the calculations later on divides by this
         variable, set it to 1 if it happens to be zero */
      d=(int)catwin[sl]; if (!d) d++;

      /* Work out the percentage share of the total prize pool that each
         tier gets */
      switch (sl)
      {
         case 4: p=22; break; /* 4-match = 22% of prize pool */
         case 5: p=10; break; /* 5-match = 10% */
         case 6: p=52;        /* Jackpot = 52% */
                 if (!have_bonus_ball ||
                     (sofar>drawn && !catwin[sl+1] && catwin[sl]))
                    p+=16; /* Throw in 5+B slice if not won and JP is won */
                 if (rollover)
                 {
                    if (topup) roll=rollover; else exact=rollover;
                 }
                 break;
         case 7: p=16; break; /* 5+bonus = 16% */
      }

      if (sl>=min_match)
      {
         /* At this point, we're in a winning table row, so update the
            running total of prize winners and the total prize pool. */
         if (overflow)
         {
            /* Overflow in operation - all tiers get equal share to start */
            przwin[sl]=overflow;
            /* If it's the jackpot tier, then they additionally get an
               equal share of any rollover/SD (rounded down to nearest
               pound, presumably) */
            if ((sl==6) && rollover) przwin[sl]+=(long)(rollover/d);
         }
         else
         if (sl>min_match)
         {
            /* If we've not got a fixed prize (3 match = 10 pounds), calculate
               individual prize according to the appropriate formula */
            if (exact) przwin[sl]=(long)(exact/d); else
            przwin[sl]=(long)((z*p+100.0*roll)/(100.0*d));
         }
      }
   }

   /* combidx[] returns to the fold */
   restore_ball_set();

   if (timings)
   {
      /* Finish timing live draw */
      double live_time=end_micro_time("table");
      if (live_time<min_time) { min_time=live_time; min_time_ball=sofar; }
      if (live_time>max_time) { max_time=live_time; max_time_ball=sofar; }
      tot_time+=live_time;
   }

   if (csv_output) show_csv_results(sofar); else show_formatted_results(sofar);
}

void show_number_results(int num_numbers, int last_num_drawn)
{
    int y;
    /* You may be amused to note that at this point, I've already thrown away
       the numbers in drawn order. Probably not too wise, since the TV doesn't
       sort them until the bonus number is drawn, hmmm... */
    (void)fprintf(stderr,"\nNumbers in ascending order: ");
    for (y=1;y<=num_numbers;y++)
    if (y>drawn) (void)fprintf(stderr,"(bonus %2d)",combidx[y]);
    else (void)fprintf(stderr,"%2d ",combidx[y]);
    (void)fprintf(stderr,"\n\n");
      
    /* May as well show how many tickets had the most recent number drawn -
       a nice feature we can provide at no cost really. */
    (void)fprintf(stderr,"%s ticket%s contained the last drawn number %d\n\n",
                  mills(num_picked[last_num_drawn]),
                  (last_num_drawn==1?"":"s"),last_num_drawn);

    /* Calculate winners/prizes table for 'num_numbers' balls so far */
    display_calcs(num_numbers);

    /* Note that at this point, any child threads should all have
       exited, having completed their jobs. */
}

void reset_fake_counts()
{
   /* No faked jobs yet, so zero highest slice/running totals
      for match 1 to 4 jobs */
   int ml;
   for (ml=0;ml<=totdrawn;ml++) { high_comb[ml]=0; fake_comb[ml]=0; }
}

void reset_saved_cats()
/* No saved match 1-4 figures can be saved from the 6th number table yet */
{
   int zsc;
   for (zsc=0;zsc<=totdrawn;zsc++) saved_catwin[zsc]=-1;
}

void monitor_file()
/* Keep an eye on a file for freshly drawn numbers. If multi-threading,
   then also do "fake" drawings of numbers remaining in the pot, so that
   some or all of the match 1-4 calculations can be performed before the
   "real" ball is drawn out. */
{
   FILE *inhand;
   struct stat finfo;
   int oldnumsofar=0,lastnumdrawn=0;
   time_t oldtimestamp=0,timestamp;
   long oldfilesize=0,filesize;
   int thenum=-1,firstloop=1;
   double last_sleep=micro_time();

#ifdef MULTITHREAD
   /* Simulate drawing each of the remaining balls and
      find out how many category winners in match 1-4 there are.
      We pass 1 so that the routine will fire up threads for us too. */
   batch_fake_jobs(1);
#endif

   while (thenum || numsofar<=drawn)
   {
      /* Don't sleep the first time in the loop */
      if (firstloop) firstloop=0;
      else
      {
         /* Find out if there's any sleep time left to take,
            usleep() it and then record when the sleep finished */
         double next_sleep=micro_time();
         double sleep_left=FILE_SLEEP-(next_sleep-last_sleep);
         if (sleep_left>0)
         {
            last_sleep=next_sleep+sleep_left;
#ifdef MULTITHREAD
            if (cores_used>1 && numjobs>0 && !all_jobs_done)
            {
               PTR_CAST_TYPE parent_thread_num=0;
               parent_time_limit=last_sleep;
               (void)thread_handler((void *)parent_thread_num);
            }
            else
#endif
            usleep(1000000.0*sleep_left);
         } else last_sleep=next_sleep;
      }

      numsofar=0;
      if (!stat(drawn_file,&finfo))
      { filesize=finfo.st_size; timestamp=finfo.st_mtime; }
      else
      { filesize=0; timestamp=0; }

      if (filesize>0 && timestamp>0 &&
          (filesize!=oldfilesize || timestamp!=oldtimestamp) &&
          (inhand=fopen(drawn_file,"r"))!=(FILE *)NULL)
      {
         char inbuff[BUFSIZ];
         int thesame=0;

         while ((thenum || numsofar<=drawn) &&
                 fgets(inbuff,BUFSIZ,inhand)!=(char *)NULL)
         {
            int bufpos=0;
            while (sscanf(&inbuff[bufpos],"%d",&thenum)!=EOF &&
                   (thenum || numsofar<=drawn))
            {
               /* Got a non-zero number, so add it to the set.
                  Can be negative to delete earlier numbers. */
               if (add_next_number(thenum))
               {
                  lastnumdrawn=thenum;
                  /* Keep things in sorted order by removing the new
                     number from the end of the array and putting it
                     back again in its ascending order position. */
                  (void)insert_number(numsofar-1,thenum);
               }

               /* Skip past everything up to next digit (or EOL) */
               while (inbuff[bufpos]!='\0' &&
                      (inbuff[bufpos]<'0' || inbuff[bufpos]>'9')) bufpos++;

               /* Skip past all digits */
               while (inbuff[bufpos]>='0' && inbuff[bufpos]<='9') bufpos++;
            }
         }
         (void)fclose(inhand);

         if (numsofar && thenum)
         {
            if (numsofar==oldnumsofar)
            {
               int cl;
               for (cl=1;(cl=numsofar) && !thesame;cl++)
               thesame=(combidx[cl]==saveidx[cl]);
            }

            if (!thesame)
            {
#ifdef MULTITHREAD
               stop_fake_jobs(lastnumdrawn);
               wait_for_jobs=0; numjobs=0; more_jobs_to_come=1;
               all_jobs_done=0;
               if (cores_used>1) spawn_threads(numsofar);
#endif
               /* Reset saved match 1-4 figures unless we're doing
                  the bonus number */
               if (numsofar<=drawn) reset_saved_cats();

               show_number_results(numsofar,lastnumdrawn);
#ifdef MULTITHREAD
               /* Fire up threads/jobs again if not all balls done yet */
               if (numsofar<=drawn) batch_fake_jobs(1);
#endif
            }
         }
      }

      save_ball_set();
      oldnumsofar=numsofar;
      oldfilesize=filesize;
      oldtimestamp=timestamp;
   }
}

void perform_draw(void)
/* Do the live draw of 7 balls */
{
   /* Loop variables. endx will include an "extra" iteration if
      7 valid -b params haven't been supplied. That iteration is
      used to read a "0" terminator or "-1" to delete the bonus ball. */
   int x=1,endx=totdrawn+(numsofar<totdrawn);

   /* No fake jobs yet */
   reset_fake_counts();

   reset_saved_cats(); /* Haven't had 6th number table yet */

   /* May need to look at a file rather than -b/keyboard */
   if (draw_from_file) monitor_file();
   else
   /* -b or keyboard input - do each number in turn */
   while (x<=endx)
   {
      int gotnum=0,n;

      /* No saved match 1-4 figures from 6th number table yet */
      if (x<=drawn) reset_saved_cats();

#ifdef MULTITHREAD
      /* Fire up those child thread puppies if we've asked for them.
         We could actually spawn them later on, but there's potential
         keyboard input coming up for the next ball that's drawn live,
         which allows me the excuse of setting up the threads *before*
         the clock timing starts :-)
         Note that we subtract one from the loop variable because the
         number hasn't been input yet. No threads are spawned if we're
         in the zero-terminator loop iteration */
      if (cores_used>1 && x<=totdrawn) spawn_threads(x-1);
#endif

      while (!gotnum)
      {
         /* May have been supplied by multiple -b params already of course,
            so try to fill those in first if any are available (and valid!) */
         if (x<=numsofar && insert_number(x-1,(n=combidx[x]))) gotnum=n;
         else
         {
            /* Temporary buff for prompt */
            char buff[BUFSIZ];
            /* Still need some more drawn numbers, so ask for them */
            if (x<=drawn)
               (void)sprintf(buff,"Input drawn number %d of %d%s",x,drawn,
                             (x>1?" (or -1 to delete previous one)":""));
            else
            if (x==totdrawn) (void)sprintf(buff,"Input bonus number (or -1 to delete 6th main number)");
            else (void)sprintf(buff,"Input 0 to exit (or -1 to delete %s number)",
                               (have_bonus_ball?"bonus":"6th main"));
#ifdef MULTITHREAD
            if (cores_used>1 && x<=totdrawn)
            {
               /* Start up some "fake" child thread jobs if they aren't already going.
                  Unfortunately, the parent process can't join in because it's got to get
                  interactive input from the user. Note that we pass zero to indicate
                  that threads are already spawned and don't need firing up. */
               if (!numjobs) batch_fake_jobs(0);
            }
            else
            /* Human asleep at keyboard with no jobs queued, so make child threads usleep() */
            wait_for_jobs=1;
#endif
            if (!(n=(int)ask_for_value(buff)))
            {
               if (x<=totdrawn)
                  (void)fprintf(stderr,"Nothing read from keyboard - try again\n");
               else gotnum=1;
            }
            else
            if (n<0)
            {
#ifdef MULTITHREAD
               stop_fake_jobs(0); /* Halt any faked jobs */               
#endif
               x+=n; if (x<1) x=1; /* Go back some balls */
               numsofar=x-1; /* Adjust global ball count */
#ifdef MULTITHREAD
               if (cores_used>1 && x<=totdrawn) spawn_threads(x-1); /* Restart threads */
#endif
            }
            else
            if (x>totdrawn)
               (void)fprintf(stderr,"You cannot enter a positive number at this point - try again\n");
            else
            if (insert_number(x-1,n))
            {
               gotnum=n;
#ifdef MULTITHREAD
               stop_fake_jobs(n); /* Halt any faked jobs */             
               /* Human woke up and typed a good number in, so tell child
                  threads to avoid usleep(), there's a-jobs a-comin' */
               wait_for_jobs=0;
#endif
            }
         }
      }

#ifdef MULTITHREAD
      /* Need to restart threads here if the fake jobs have been run.
         This will include the latest number in any "real" jobs later on. */
      if (cores_used>1 && x<=totdrawn) spawn_threads(x);
#endif
 
      /* Show results after number is drawn */
      if (x<=totdrawn) show_number_results(x,gotnum);
      x++;
   }

   if (timings)
   {
      /* Display summary of table generation timings */
      (void)fprintf(stderr,"Minimum table gen time =  %0.06f seconds (for ball %d)\n",
                    min_time,min_time_ball);
      (void)fprintf(stderr,"Maximum table gen time =  %0.06f seconds (for ball %d)\n",
                    max_time,max_time_ball);
      (void)fprintf(stderr,"Total   table gen time =  %0.06f seconds (for %d balls)\n",
                    tot_time,totdrawn);
      (void)fprintf(stderr,"Average table gen time =  %0.06f seconds\n\n",
                    tot_time/totdrawn);
   }
}
