/* pthread.c - Code to handle multi-threading.
   (C) Richard K. Lloyd 2010 <webmaster@lottery.merseyworld.com> */

/* Get the usual definitions */
#include "draw.h"

/* Variables and functions declared in other source files */
extern int cores_used;
extern int high_comb[MAX_DRAWN+1],fake_comb[MAX_DRAWN+1];
extern double micro_time(void),parent_time_limit;
extern int do_combo_matches(int thread_number,int sofar,int sl,int inserted_ball,int start_point,int end_point);
extern void tidy_up(void),reset_fake_counts();

/* Allocate thread structures for up to MAX_CPU_CORES-1 threads (index 0
   is for the parent process, but it doesn't use it because it's not a
   thread :-) ) */
static pthread_t threads[MAX_CPU_CORES];

/* Exit system time for a thread (<0.0 = stay in thread until all jobs done) */
double thread_time_limit[MAX_CPU_CORES];

/* Any child threads at all alive? */
static int kids_alive=0;

/* All jobs completed (i.e. don't enter thread_handler any more) */
int all_jobs_done=0;

/* Indicate thread running status:
            -2 = Running, but thread cancellation requested
            -1 = Running, but no job allocated
             0 = Not running
   1...numjobs = Running that job

   Note that we malloc num_cores*sizeof(int) bytes to this
   pointer to create a memory array that the threads will index into. */
int *thread_running=(int *)NULL;

/* Job status codes for threads (stored in job_status[1..MAX_JOBS] array):
   0-MAX_CPU_CORES-1 = CPU core 0..MAX_CPU_CORES-1 has been allocated job
                       (note that "CPU core 0" is actually the parent process)
   MAX_CPU_CORES     = Idle, no job ready for allocation yet
   MAX_CPU_CORES+1   = Idle, job ready to be allocated
   MAX_CPU_CORES+2   = Idle, job completed. */
int job_status[MAX_JOBS+1];

/* Copy of job_status, but just keeps the last value of 0..MAX_CPU_CORES-1
   (i.e. never gets cleared to MAX_CPU_CORES+2 when job is completed). It's
   mainly for debugging so we know which thread ran which job even after
   all jobs/threads are long since done/stopped. */
int job_thread[MAX_JOBS+1];

/* Number of matches the job is processing (1-4) */
int job_matches[MAX_JOBS+1];

/* "Fake" number inserted into drawn balls for job (0 = no fake number) */
int fake_ball[MAX_JOBS+1];

/* Start number of thread_combidx[1] for combination comparison */
int start_combo_num[MAX_JOBS+1];

/* End number of thread_combidx[1] for combination comparison */
int end_combo_num[MAX_JOBS+1];

/* Return the number of matches in this array. */
int match_counts[MAX_JOBS+1];

/* Number of balls drawn so far */
int thread_sofar=0;

/* Number of jobs allocated by the parent process so far */
int numjobs;

/* Flag to indicate whether more jobs are to come on the queue.
   No child thread can exit its work loop unless more_jobs_to_come=0. */
int more_jobs_to_come=1;

/* Do child threads need to usleep() for jobs (=1) or have all
   the balls been supplied via -b already (=0)? */
int wait_for_jobs=1;

/* Forward declarations to keep some compilers quiet */
void *thread_handler(void *threadptr);
void spawn_threads(int balls_sofar);

static void hammer_cpu(int hammer_units)
/* Hammer the CPU for the specified number of HAMMER_TIME*hammer_units seconds.
   It's a crude way of telling the OS scheduler "don't allocate the next
   child thread to the same CPU core that I'm running on". There might be a way
   to allocate a thread to a particular CPU of course, but this rudimentary
   method should hopefully work. */
{
   double end_hammer=micro_time()+HAMMER_TIME*hammer_units;
   while (micro_time()<end_hammer) {}; /* This will blitz the CPU */
}

void set_affinity(int thread_number,int single_core)
/* Set CPU core affinity for thread 'thread_number' (0=parent process).
   If single_core=1, set thread/process to use CPU core 'thread_number'
   otherwise allow thread/process to run any any CPU core (0..cores_used-1).
*/
{
#ifdef HAVE_AFFINITY
   cpu_set_t the_cpu_set;
   int af;

   CPU_ZERO(&the_cpu_set); /* No CPUs in set */

   if (single_core) CPU_SET(thread_number,&the_cpu_set);
   else for (af=0;af<cores_used;af++) CPU_SET(af,&the_cpu_set);
    
   if (thread_number)
     (void)pthread_setaffinity_np(threads[thread_number],
                                  sizeof(cpu_set_t),&the_cpu_set);
   else
     (void)sched_setaffinity(0,sizeof(cpu_set_t),&the_cpu_set);
#endif
}

void *thread_handler(void *threadptr)
/* This is the routine that each spawned thread spends its entire life in.
   OK, it can call other functions in this binary, but we want to avoid
   all external calls - except to pthread library routines naturally enough -
   as possible because of the issue of thread safety (and the simple nasty
   mixing of stuff from different threads). Yes, this even includes not
   using something as basic as fprintf(). */
{
   /* What each thread needs to do in a loop:
      - Look for the next "job" on the queue:
        - If none placed yet, sleep and check again later.
        - If all jobs done, exit.
      - Run any found "job" with this input:
        * How many numbers have been drawn so far (1-7).
        * Numbers drawn so far in ascending order.
        * Whether Match 1-4 is to be calculated.
        * Slice of combinations needed (first number range).
      - Flag job is complete when it returns with:
        * Number of matches for Match 1-4 for that combination slice.
   */

   /* Thread number (parent process = 0) */
   PTR_CAST_TYPE thread_num=(PTR_CAST_TYPE)threadptr;
   int jc,jobsdone=0,high_numjobs=numjobs,time_limit_reached=0;
   double thread_time_limit;

   /* Flag that we're now running a parent or child thread
      (-1 = no job yet) */
   thread_running[thread_num]=-1;

   if (thread_num)
   {
      /* Force this child thread onto CPU core thread_num
         (CPU cores are numbered 0..cores_used-1). Remember that the parent
         process will later be forced to CPU core 0. */
      set_affinity(thread_num,1);

      /* No time limit for child threads */
      thread_time_limit=-1.0;

      /* If it's not the final child thread to be allocated and it's not the
         parent CPU (thread_num=0), then blitz the CPU so that later child
         threads will hopefully avoid the CPU this thread is running on. Time
         the CPU hammering to finish just after the final child thread is
         allocated. */
      if (thread_num<(long)cores_used-1)
         hammer_cpu((int)(cores_used-thread_num-1));
   } else thread_time_limit=parent_time_limit;

#ifdef DEBUG
   (void)fprintf(stderr,"Thread #%d spawned OK (numjobs=%d)\n",
                 (int)thread_num,high_numjobs);
#endif

   /* No point even looking for jobs if there aren't any.
      Note that if all the balls have been supplied via -b (or any remaining
      balls have been typed in), then we don't need to usleep() in the job
      search loop, because the jobs are "imminent" and we want to start
      running them as soon as it is possible. */
   while (!numjobs && more_jobs_to_come && !all_jobs_done &&
          thread_running[thread_num]==-1)
   if (wait_for_jobs) (void)usleep(THREAD_SLEEP);

   /* A tiny bit of sleeping is needed here (don't ask me why!) */
   if (thread_running[thread_num]==-1) usleep(THREAD_SLEEP/10);

#ifdef DEBUG
   (void)fprintf(stderr,"Thread #%d entering job processing loop (numjobs=%d)\n",
                 (int)thread_num,high_numjobs);
#endif

   /* Keep a track of highest number of jobs at any time */
   if (numjobs>high_numjobs) high_numjobs=numjobs;

   /* We now have at least one job on the queue, so Keep checking until
      all the jobs are done. If more jobs are still to be put on the
      queue, don't leave the scanning loop. */
   while (thread_running[thread_num]!=-2 && numjobs && !time_limit_reached &&
          (jobsdone<numjobs || more_jobs_to_come))
   {
      /* Looking for a job to run to calculate "chunks" of the matches 1..4
         combinations */
 
#ifdef DEBUG
      (void)fprintf(stderr,"Thread #%d in inner job loop (more_jobs=%d,jobsdone=%d,numjobs=%d)\n",
                    (int)thread_num,more_jobs_to_come,jobsdone,high_numjobs);
#endif

      jobsdone=0; jc=0;

      /* Keep a track of highest number of jobs at any time */
      if (numjobs>high_numjobs) high_numjobs=numjobs;

      while (thread_running[thread_num]!=-2 && jc<numjobs && !time_limit_reached)
      switch (job_status[++jc])
      {
         case IDLE_NO_JOB_READY: /* No job allocated in this slot
                                    (shouldn't happen, so ignore it anyway) */
            break;

         case IDLE_JOB_READY: /* Job ready for assignment */

          /* Keep a track of highest number of jobs at any time */
          if (numjobs>high_numjobs) high_numjobs=numjobs;

         job_status[jc]=(int)thread_num; /* Try and assign our thread to it */

         /* Although checked for a job and found one *and* also assigned our
            thread to it, another thread might have snuck in and grabbed it
            inbetween the check and the assignment, so check we still have it
         */
         if (job_status[jc]==thread_num)
         {
            int mcount=0;
            /* Yep, we're now totally sure we've been assigned the job, so
               run it. */

            if (numjobs && (thread_running[thread_num]==-1))
            {
               thread_running[thread_num]=jc; /* Use job number for thread flag */
               job_thread[jc]=thread_num; /* Never gets cleared */
#ifdef DEBUG
               (void)fprintf(stderr,"Allocated match %d job #%d (%d balls so far) to thread #%d\n",
                             job_matches[jc],jc,thread_sofar,(int)thread_num);
#endif
            }

            if (numjobs && (thread_running[thread_num]==jc))
            mcount=do_combo_matches(thread_num,thread_sofar,
                                    job_matches[jc],fake_ball[jc],
                                    start_combo_num[jc],end_combo_num[jc]);

            /* If jobs terminated early (numjobs=0), don't count return value
               of do_combo_matches() because may have exited it early. */
            if (numjobs && (thread_running[thread_num]==jc))
            {
               match_counts[jc]=mcount;
               job_status[jc]=IDLE_JOB_COMPLETED; /* And we're done */
               jobsdone++; /* And, yes, increment the job count */
#ifdef DEBUG
               (void)fprintf(stderr,"Thread %d: m=%d/j=%d/s=%d/f=%d/st=%d/en=%d/nj=%d job returned mt=%d\n",
                             (int)thread_num,job_matches[jc],jc,thread_sofar,
                             fake_ball[jc],
                             start_combo_num[jc],end_combo_num[jc],high_numjobs,
                             mcount);
#endif
            }
            else job_status[jc]=IDLE_JOB_READY; /* Didn't finish job */

            if (thread_running[thread_num]!=-2)
               thread_running[thread_num]=-1; /* No job allocated now */

            /* If it's the parent process, we need to see if we've hit the
               time limit yet */
            if (thread_time_limit>0.0 && micro_time()>=thread_time_limit)
               time_limit_reached=1;
         } else jobsdone++; /* Another thread "stole" our job */
         break;

         case IDLE_JOB_COMPLETED: /* Job completed already */
           jobsdone++; /* Another job done, so increment count */
           break;

         default: /* Everything else means another thread is running job */
           jobsdone++; /* We can't run it ourselves, so consider job done */
           break;
      }
   }

   /* Flag if all jobs are done */
   if (jobsdone==high_numjobs) all_jobs_done=1;
     
#ifdef DEBUG
   (void)fprintf(stderr,"Thread #%d exited %s (%d/%d jobs done)\n",
                 (int)thread_num,
                 ((!numjobs || thread_running[thread_num]==-2)?"early":"normally"),
                 jobsdone,high_numjobs);
#endif

   thread_running[thread_num]=0; /* Yes, it's death to child */
   if (thread_num) pthread_exit(NULL); /* Kill child thread */
   return((void *)NULL); /* Keep gcc happy and continue parent process */
}

void cancel_thread(int tnum)
/* Cancel thread number 'tnum' if it's still running */
{
   if (tnum>0 && tnum<cores_used &&
       thread_running[tnum]!=-2 && thread_running[tnum]!=0)
   {
#ifdef DEBUG
      (void)fprintf(stderr,"Flagged thread #%d to exit\n",tnum);
#endif
      thread_running[tnum]=-2;
   }
}

void kill_all_children()
/* Tell child threads to exit immediately */
{
   if (kids_alive)
   {
      int tl;
      double runfrom=micro_time();
      /* Only hang around for 0.1 seconds */
      double runsofar=runfrom,runto=runfrom+0.1;

      thread_running[0]=0; /* Parent is running this code, so it's "dead" :-) */

      /* Flag that there's no jobs now or to come */
      numjobs=0; more_jobs_to_come=0; all_jobs_done=1;

      /* Kill those children now */
      for (tl=1;tl<cores_used;tl++) cancel_thread(tl);

      for (tl=1;tl<cores_used && runsofar<runto;tl++)
      while (((runsofar=micro_time())<runto) && thread_running[tl]) { }

      kids_alive=0; /* All children are dead now */
   }
}

void spawn_threads(int balls_sofar)
{
   PTR_CAST_TYPE tl;

   /* Kill all running child threads first */
   kill_all_children();

   /* Globally set number of balls drawn so far */
   thread_sofar=balls_sofar;

   /* Initially use all CPU cores for parent process */
   set_affinity(0,0);

   /* Tell everyone we're sitting around doing absolutely nothing */
   for (tl=1;tl<=MAX_JOBS;tl++)
   job_status[tl]=IDLE_NO_JOB_READY; /* Idle, no jobs ready for allocation */

   /* No jobs yet, but there will be some soon */
   more_jobs_to_come=1; all_jobs_done=0;

   for (tl=1;tl<cores_used;tl++)
   {
      /* Hammer parent process CPU, so CPU scheduler won't use it when
         the child process is created - well, that's the theory anyway. */
      hammer_cpu(1);

      if (pthread_create(&threads[tl],NULL,thread_handler,(void *)tl))
      {
         (void)fprintf(stderr,"Can't spawn child thread #%d - aborted\n",
                       (int)tl);
         tidy_up();
         exit(1);
      }
      kids_alive=1; /* They're alive! */
   }

   set_affinity(0,1); /* Now force parent process to use CPU 0 */
}

void put_jobs_on_queue(int num_matches,int faked_ball)
/* Put a job on the queue with the following attributes:
   num_matches = Number of matches (1-4)
   faked_ball  = Fake ball inserted into drawn balls (0 = no insertion)

   Note: Only call this function if MULTITHREAD is defined and
         cores_used>1.
*/
{
   /* We're multi-threading, using more than 1 CPU core, so it's
      time to put some match 1-4 jobs onto the queue. */

   int slice_end=0,
       slices_done=(high_comb[num_matches]==upper-drawn+1),
       new_numjobs=numjobs;

   /* If we're trying to do the bonus number, we can't fake the ball */
   if (thread_sofar>=drawn) faked_ball=0;

   /* Set up start and end slice for jobs here.
      Could be a nasty problem if numjobs ends up > MAX_JOBS,
      so check for that. */

   while (!slices_done)
   {
      int slice_start=slice_end+1,slice_size=0;

      if (new_numjobs==MAX_JOBS)
      {
         (void)fprintf(stderr,"Too many jobs to allocate (MAX_JOBS=%d, which is too small) - aborted\n",MAX_JOBS);
         tidy_up();
         exit(1);
      }
      else
      if (slice_start<=upper-drawn+1)
      {
         slice_end=slice_start;
         /* Split 14m combos into smaller slices */
         while (!slices_done &&
                slice_size+combo_slice[slice_end]<scaled_speed_combo[num_matches])
         {
            slice_size+=combo_slice[slice_end];
            if (slice_end<upper-drawn+1) slice_end++;
            else slices_done=1;
         }

         /* Only do this slice if we haven't faked it already */
         if (slice_start>high_comb[num_matches])
         {
            /* Possibly insert fake ball */
            fake_ball[++new_numjobs]=faked_ball;

            /* Mark the start and end point of the slice for job */
            start_combo_num[new_numjobs]=slice_start;
            end_combo_num[new_numjobs]=slice_end;

            /* Store number of matches job will use */
            job_matches[new_numjobs]=num_matches;

            /* No category matches yet */
            match_counts[new_numjobs]=0;

            /* Threads can safely grab job now */
            job_status[new_numjobs]=IDLE_JOB_READY;
#ifdef DEBUG
            (void)fprintf(stderr,"Put job #%d (sofar %d, match %d, faked %d, slice %d-%d) job on queue\n",
                          new_numjobs,thread_sofar,num_matches,faked_ball,
                          slice_start,slice_end);
#endif
         }
      } else slices_done=1;
   }

#ifdef DEBUG
   (void)fprintf(stderr,"Made %d %s jobs (number=%d) for match %d (faked already: 1-%d/%d)\n",
                 new_numjobs-numjobs,(faked_ball?"faked":"real"),faked_ball,num_matches,
                 high_comb[num_matches],upper-drawn+1);
#endif

   /* Let the children loose on the batched jobs */
   numjobs=new_numjobs;
}

void batch_fake_jobs(int fire_up_threads)
/* Batch up "fake" jobs for match 1-4 scans of
   the remaining balls left in the pot. The loops are
   ordered so that we're fairly sure that we'll get
   at least one of match scans for each of the remaining
   balls done within the 6 seconds or so we've got.
   fire_up_threads=1 means that we need to spawn threads as well. */
{
   if (cores_used>1)
   {
      int bl,fn;

      reset_fake_counts(); /* No faked counts/combs yet */
      more_jobs_to_come=1; /* We've got jobs coming down the line */
      if (fire_up_threads)
      {
         wait_for_jobs=0;
         spawn_threads(numsofar);
      }

      /* Only do fake jobs if either reading from a file
         (which can't guarantee that we've got the 6th ball's
         match 1-4 figures) or not doing bonus ball faking */
      if (thread_sofar<drawn || fire_up_threads)
      for (bl=0;bl<totdrawn;bl++)
      {
         int nm=match_order[bl];
         if (nm>=1 && nm<=4 && nm<=thread_sofar+1)
         for (fn=1;fn<=upper;fn++)
         {
            int gf=0,il;
            for (il=1;il<=thread_sofar && !gf;il++)
            gf=(fn==saveidx[il]);
            if (!gf) put_jobs_on_queue(nm,fn);
         }
      }
      more_jobs_to_come=0; /* No more on queue */
   }
}

void stop_fake_jobs(int latestnum)
/* See how far the fake jobs got...
   If latestnum=0, don't bother totalling the counts */
{
   if (cores_used>1)
   {
      int old_numjobs=numjobs,jl,
          gotwork=(latestnum>0 && numjobs>0),finished_jobs=0;
      kill_all_children(); /* Stop child threads ASAP */

      /* Count up the match totals and highest combo slice scanned
         for the faked jobs that were for the "real" number drawn. */
      while (gotwork)
      {
         gotwork=0;
         for (jl=1;jl<=old_numjobs;jl++)
         {
            int mt=job_matches[jl];
            if (job_status[jl]==IDLE_JOB_COMPLETED &&
                fake_ball[jl]==latestnum &&
                start_combo_num[jl]==high_comb[mt]+1)
            {
               high_comb[mt]=end_combo_num[jl];
               fake_comb[mt]+=match_counts[jl];
               /* Make sure not re-used by resetting job status */
               job_status[jl]=IDLE_NO_JOB_READY;
               gotwork=1; finished_jobs++;
#ifdef DEBUG
               (void)fprintf(stderr,"Exact fake job #%d: match %d, slice %d-%d, wins=%d, win_tot=%d\n",
                             finished_jobs,mt,start_combo_num[jl],high_comb[mt],match_counts[jl],fake_comb[mt]);
#endif
            }
         }
      }

#ifdef DEBUG
      if (latestnum)
      {
         for (jl=1;jl<=4 && jl<=thread_sofar+1;jl++)
         (void)fprintf(stderr,"Fakeres = sofar %d, match %d, wins %d, sliced %d\n",
                       thread_sofar,jl,fake_comb[jl],high_comb[jl]);
         (void)fprintf(stderr,"Completed %d fake background jobs for real drawn number %d\n",
                       finished_jobs,latestnum);
      }
#endif
   }
}
